#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""

Remote Service Admin API

:author: Scott Lewis
:copyright: Copyright 2020, Scott Lewis
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Scott Lewis

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from datetime import datetime
from distutils.util import strtobool
from traceback import print_exception
import logging
import threading
import sys

# Typing
try:
    # pylint: disable=W0611
    from typing import Any, Dict, List, Optional, Tuple
    from pelix.framework import Bundle, BundleContext
    from pelix.internals.registry import ServiceRegistration
    from pelix.rsa.providers.distribution import (
        ExportContainer,
        ImportContainer,
    )
except ImportError:
    pass

from pelix import constants
from pelix.constants import (
    BundleActivator,
    SERVICE_RANKING,
    OBJECTCLASS,
    OSGI_FRAMEWORK_UUID,
)
from pelix.framework import BundleException
from pelix.internals.registry import ServiceReference

from pelix.ipopo.decorators import (
    ComponentFactory,
    Provides,
    Instantiate,
    Validate,
    Invalidate,
    Requires,
    RequiresBest,
)

from pelix.rsa.edef import EDEFWriter
from pelix.rsa.endpointdescription import EndpointDescription

from pelix.rsa import (
    SelectImporterError,
    RemoteServiceError,
    validate_exported_interfaces,
    RemoteServiceAdminEvent,
    SERVICE_REMOTE_SERVICE_ADMIN,
    SERVICE_RSA_EVENT_LISTENER,
    get_exported_interfaces,
    SERVICE_EXPORTED_INTERFACES,
    get_edef_props_error,
    REMOTE_CONFIGS_SUPPORTED,
    SERVICE_EXPORTED_CONFIGS,
    SERVICE_INTENTS,
    SERVICE_EXPORTED_INTENTS,
    SERVICE_EXPORTED_INTENTS_EXTRA,
    ECF_ENDPOINT_TIMESTAMP,
    ENDPOINT_FRAMEWORK_UUID,
    get_current_time_millis,
    get_string_plus_property,
    ExportReference,
    ExportRegistration,
    set_append,
    ImportReference,
    ImportRegistration,
    RemoteServiceAdminListener,
)

from pelix.rsa.providers.distribution import (
    SERVICE_EXPORT_DISTRIBUTION_PROVIDER,
    SERVICE_IMPORT_DISTRIBUTION_PROVIDER,
)

# ------------------------------------------------------------------------------
# Module version

__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# Standard logging
_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------

# Framework property that allows RSA debug to be disabled. To disable automatic
# output of RSA events, set the property 'pelix.rsa.remoteserviceadmin.debug
# to some string other than 'true' (the default)
DEBUG_PROPERTY = "pelix.rsa.remoteserviceadmin.debug"
DEBUG_PROPERTY_DEFAULT = "false"

# ------------------------------------------------------------------------------


@BundleActivator
class Activator(object):
    """
    Bundle activator. By default, register an instance of
    RemoteServiceAdminEventListener service for debugging.
    This can be disabled by setting the DEBUG_PROPERTY aka
    ``pelix.rsa.remoteserviceadmin.debug`` to ``false``.
    """

    def __init__(self):
        self._context = None
        self._debug_reg = None

    def start(self, context):
        """
        Bundle starting
        """
        self._context = context
        debug_str = self._context.get_property(DEBUG_PROPERTY)
        if not debug_str:
            debug_str = DEBUG_PROPERTY_DEFAULT

        if strtobool(debug_str):
            self._debug_reg = self._context.register_service(
                SERVICE_RSA_EVENT_LISTENER,
                DebugRemoteServiceAdminListener(),
                None,
            )

    def stop(self, _):
        """
        Bundle stopping
        """
        if self._debug_reg:
            self._debug_reg.unregister()
            self._debug_reg = None
        self._context = None


# ------------------------------------------------------------------------------


SERVICE_EXPORT_CONTAINER_SELECTOR = "pelix.rsa.exportcontainerselector"


class ExportContainerSelector:
    """
    RSA Impl Export Container Selector service specification and default
    implementation.  The highest priority instance of this service available
    at runtime is used by the RemoteServiceAdmin implementation to select
    export containers to handle a given call to
    RemoteServiceAdmin.export_service
    """

    # pylint: disable=R0903
    def select_export_containers(
        self, service_ref, exported_intfs, export_props
    ):
        # type: (ServiceReference, List[str], Dict[str, Any]) -> List[ExportContainer]
        # pylint: disable=W0613
        """
        Select export containers, given service_ref (ServiceReference),
        exported_refs list(string), and export_props dict(string:?).  Each of
        these parameters must not be None.

        :param service_ref ServiceReference for possible export
        :param exported_intf list of strings defining the service specifications
        to be exported for the service
        :param export_prop the service properties to be used for export after
        combining the service_ref properties with any overriding properties.
        :return list of ExportContainer instances that will be responsible for
        exporting the given service.  Will not return None, but may return
        empty list
        """
        raise Exception(
            "{0}.select_export_containers is not implemented".format(self)
        )


@ComponentFactory("pelix-rsa-exporterselector-factory")
@Provides(SERVICE_EXPORT_CONTAINER_SELECTOR)
@Requires(
    "_export_distribution_providers",
    SERVICE_EXPORT_DISTRIBUTION_PROVIDER,
    True,
    True,
)
@Instantiate(SERVICE_EXPORT_CONTAINER_SELECTOR, {SERVICE_RANKING: -1000000000})
class ExportContainerSelectorImpl(ExportContainerSelector):
    def __init__(self):
        self._export_distribution_providers = []

    def select_export_containers(
        self, service_ref, exported_intfs, export_props
    ):
        # type: (ServiceReference, List[str], Dict[str, Any]) -> List[ExportContainer]
        # get exported configs
        exported_configs = get_string_plus_property(
            SERVICE_EXPORTED_CONFIGS, export_props, None
        )
        # get service intents, via service.intents, services.exported.intents,
        # and extra
        service_intents_set = set_append(
            set(), export_props.get(SERVICE_INTENTS, None)
        )
        service_intents_set = set_append(
            service_intents_set,
            export_props.get(SERVICE_EXPORTED_INTENTS, None),
        )
        service_intents_set = set_append(
            service_intents_set,
            export_props.get(SERVICE_EXPORTED_INTENTS_EXTRA, None),
        )

        export_containers = []
        for export_provider in self._export_distribution_providers:
            export_container = export_provider.supports_export(
                exported_configs, list(service_intents_set), export_props
            )
            if export_container:
                export_containers.append(export_container)

        return export_containers


# ------------------------------------------------------------------------------
# Import Container Selector service specification and default
# implementation.  The highest priority instance of this service available
# at runtime is used by the RemoteServiceAdmin implementation to select
# a single import container to handle a given call to RemoteServiceAdmin.import_service
SERVICE_IMPORT_CONTAINER_SELECTOR = "pelix.rsa.importcontainerselector"


class ImportContainerSelector:
    # pylint: disable=R0903
    def select_import_container(self, remote_configs, endpoint_description):
        # type: (List[str], EndpointDescription) -> ImportContainer
        # pylint: disable=W0613
        """
        Select import container, given endpoint_description
        (EndpointDescription).
        Returns a single ImportContainer, or None.

        :param endpoint_description: EndpointDescription describing endpoint for
                                     possible import
        :return ImportContainer instance or None
        """
        raise Exception(
            "{0}.select_import_container is not implemented".format(self)
        )


# ------------------------------------------------------------------------------
@ComponentFactory("pelix-rsa-importerselector-factory")
@Provides(SERVICE_IMPORT_CONTAINER_SELECTOR)
@Requires(
    "_import_distribution_providers",
    SERVICE_IMPORT_DISTRIBUTION_PROVIDER,
    True,
    True,
)
@Instantiate("pelix-rsa-importerselector-impl", {SERVICE_RANKING: -1000000000})
class ImportContainerSelectorImpl(ImportContainerSelector):
    # pylint: disable=R0903
    def __init__(self):
        self._import_distribution_providers = []

    def select_import_container(self, remote_configs, endpoint_description):
        # type: (List[str], EndpointDescription) -> ImportContainer
        for import_provider in self._import_distribution_providers:
            import_container = import_provider.supports_import(
                remote_configs,
                endpoint_description.get_intents(),
                endpoint_description.get_properties(),
            )
            if import_container:
                return import_container

        return None


# ------------------------------------------------------------------------------
# Implementation of RemoteServiceAdmin service
# ------------------------------------------------------------------------------
@ComponentFactory("pelix-rsa-remoteserviceadminimpl-factory")
@Provides(SERVICE_REMOTE_SERVICE_ADMIN)
@RequiresBest(
    "_export_container_selector", SERVICE_EXPORT_CONTAINER_SELECTOR, False
)
@RequiresBest(
    "_import_container_selector", SERVICE_IMPORT_CONTAINER_SELECTOR, False
)
@Requires("_rsa_event_listeners", SERVICE_RSA_EVENT_LISTENER, True, True)
@Instantiate("pelix-rsa-remoteserviceadminimpl")
class RemoteServiceAdminImpl(object):
    def __init__(self):
        self._context = None  # type: BundleContext
        self._exported_regs = []  # type: List[ExportRegistrationImpl]
        self._exported_regs_lock = threading.RLock()
        self._imported_regs = []  # type: List[ImportRegistrationImpl]
        self._imported_regs_lock = threading.RLock()
        self._rsa_event_listeners = []
        self._export_container_selector = None  # type: ExportContainerSelector
        self._import_container_selector = None  # type: ImportContainerSelector

    def get_exported_services(self):
        # type: () -> List[ExportReference]
        result = []
        for reg in self._get_export_regs():
            exp_ref = reg.get_export_reference()
            if exp_ref:
                result.append(exp_ref)
        return result

    def get_imported_endpoints(self):
        # type: () -> List[ImportReference]
        result = []
        for reg in self._get_import_regs():
            imp_ref = reg.get_import_reference()
            if imp_ref:
                result.append(imp_ref)
        return result

    def _get_export_regs(self):
        # type: () -> List[ExportRegistration]
        with self._exported_regs_lock:
            return self._exported_regs[:]

    def _get_import_regs(self):
        # type: () -> List[ImportRegistration]
        with self._imported_regs_lock:
            return self._imported_regs[:]

    def export_service(self, service_ref, overriding_props=None):
        # type: (ServiceReference, Dict[str, Any]) -> List[ExportRegistration]
        if not service_ref:
            raise RemoteServiceError("service_ref must not be None")
        assert isinstance(service_ref, ServiceReference)
        # get exported interfaces
        exported_intfs = get_exported_interfaces(service_ref, overriding_props)
        # must be set by service_ref or overriding_props or error
        if not exported_intfs:
            raise RemoteServiceError(
                SERVICE_EXPORTED_INTERFACES
                + " must be set in svc_ref properties or overriding_props"
            )
        # If the given exported_interfaces is not valid, then return empty list
        if not validate_exported_interfaces(
            service_ref.get_property(OBJECTCLASS), exported_intfs
        ):
            return []
        # get export props by overriding service get_reference properties
        # (if overriding_props set)
        export_props = service_ref.get_properties().copy()
        if overriding_props:
            export_props.update(overriding_props)

        # Force the framework UID, as the one from error_props
        # was generated
        export_props[ENDPOINT_FRAMEWORK_UUID] = self._context.get_property(
            OSGI_FRAMEWORK_UUID
        )

        result_regs = []
        result_events = []
        exporters = None
        error_props = get_edef_props_error(
            service_ref.get_property(OBJECTCLASS)
        )
        try:
            # get list of exporters from export_container_selector service
            exporters = self._export_container_selector.select_export_containers(
                service_ref, exported_intfs, export_props
            )
            # if none returned then report as warning at return empty list
            if not exporters:
                _logger.warning(
                    "No exporting containers found to export "
                    "service_ref=%s;export_props=%s",
                    service_ref,
                    export_props,
                )
                return []
        except:
            error_reg = ExportRegistrationImpl.fromexception(
                sys.exc_info(), EndpointDescription(service_ref, error_props)
            )
            export_event = RemoteServiceAdminEvent.fromexportreg(
                self._get_bundle(), error_reg
            )
            result_regs.append(error_reg)
            self._add_exported_service(error_reg)
            result_events.append(export_event)

        # If no errors added to result_regs then we continue
        if not result_regs:
            # get _exported_regs_lock
            with self._exported_regs_lock:
                # cycle through all exporters
                for exporter in exporters:
                    found_regs = []
                    # get exporter id
                    exporterid = exporter.get_id()
                    for reg in self._exported_regs:
                        if reg.match_sr(service_ref, exporterid):
                            found_regs.append(reg)
                    # if so then found_regs will be non-empty
                    if found_regs:
                        for found_reg in found_regs:
                            new_reg = ExportRegistrationImpl.fromreg(found_reg)
                            self._add_exported_service(new_reg)
                            result_regs.append(new_reg)
                    else:
                        # Here is where export is done
                        export_reg = None
                        export_event = None
                        ed_props = error_props

                        try:
                            # use exporter.make_endpoint_props to make endpoint
                            # props, expect dictionary in response
                            ed_props = exporter.prepare_endpoint_props(
                                exported_intfs, service_ref, export_props
                            )
                            # export service and expect and EndpointDescription
                            # instance in response
                            export_ed = exporter.export_service(
                                service_ref, ed_props
                            )
                            # if a valid export_ed was returned
                            if export_ed:
                                export_reg = ExportRegistrationImpl.fromendpoint(
                                    self, exporter, export_ed, service_ref
                                )
                                export_event = RemoteServiceAdminEvent.fromexportreg(
                                    self._get_bundle(), export_reg
                                )
                        except Exception:
                            export_reg = ExportRegistrationImpl.fromexception(
                                sys.exc_info(),
                                EndpointDescription.fromprops(ed_props),
                            )
                            export_event = RemoteServiceAdminEvent.fromexportreg(
                                self._get_bundle(), export_reg
                            )

                        # add exported reg to exported services
                        self._add_exported_service(export_reg)
                        # add to result_regs also
                        result_regs.append(export_reg)
                        # add to result_events
                        result_events.append(export_event)
            # publish events
        for e in result_events:
            self._publish_event(e)
        return result_regs

    def import_service(self, endpoint_description):
        # type: (EndpointDescription) -> ImportRegistration
        if not endpoint_description:
            raise RemoteServiceError(
                "endpoint_description param must not be empty"
            )
        assert isinstance(endpoint_description, EndpointDescription)

        remote_configs = get_string_plus_property(
            REMOTE_CONFIGS_SUPPORTED,
            endpoint_description.get_properties(),
            None,
        )
        if not remote_configs:
            raise RemoteServiceError(
                "endpoint_description must contain {0} property".format(
                    REMOTE_CONFIGS_SUPPORTED
                )
            )

        try:
            importer = self._import_container_selector.select_import_container(
                remote_configs, endpoint_description
            )
            if not importer:
                raise SelectImporterError(
                    "Could not find importer for endpoint={0}".format(
                        endpoint_description
                    )
                )
        except:
            import_reg = ImportRegistrationImpl.fromexception(
                sys.exc_info(), endpoint_description
            )
            import_event = RemoteServiceAdminEvent.fromimportreg(
                self._get_bundle(), import_reg
            )
        else:
            with self._imported_regs_lock:
                found_reg = None
                for reg in self._imported_regs:
                    if reg.match_ed(endpoint_description):
                        found_reg = reg
                        break

                if found_reg is not None:
                    # if so then found_regs will be non-empty
                    ex = found_reg.get_exception()
                    if ex:
                        new_reg = ImportRegistrationImpl.fromexception(
                            ex, endpoint_description
                        )
                    else:
                        new_reg = ImportRegistrationImpl.fromreg(found_reg)
                        new_reg.get_import_reference().update(
                            endpoint_description
                        )

                    self._add_imported_service(new_reg)
                    return new_reg

                # Here is where new import is done
                try:
                    svc_reg = importer.import_service(endpoint_description)
                    import_reg = ImportRegistrationImpl.fromendpoint(
                        self, importer, endpoint_description, svc_reg
                    )
                    import_event = RemoteServiceAdminEvent.fromimportreg(
                        self._get_bundle(), import_reg
                    )
                except:
                    import_reg = ImportRegistrationImpl.fromexception(
                        sys.exc_info(), endpoint_description
                    )
                    import_event = RemoteServiceAdminEvent.fromimportreg(
                        self._get_bundle(), import_reg
                    )

        self._imported_regs.append(import_reg)
        self._publish_event(import_event)
        return import_reg

    def _publish_event(self, event):
        listeners = self._rsa_event_listeners
        if listeners:
            for l in listeners:
                try:
                    l.remote_admin_event(event)
                except:
                    _logger.exception(
                        "Exception calling rsa event listener=%s", l
                    )

    def _get_bundle(self):
        # type: () -> Optional[Bundle]
        return self._context.get_bundle() if self._context else None

    @Validate
    def _validate(self, context):
        self._context = context

    @Invalidate
    def _invalidate(self, _):
        with self._exported_regs_lock:
            for reg in self._exported_regs:
                reg.close()
            del self._exported_regs[:]
        with self._imported_regs_lock:
            for reg in self._imported_regs:
                reg.close()

            del self._imported_regs[:]
        self._context = None

    def _unexport_service(self, svc_ref):
        # type: (ServiceReference) -> None
        with self._exported_regs_lock:
            for reg in self._exported_regs:
                if reg.match_sr(svc_ref, None):
                    reg.close()

    @staticmethod
    def _valid_exported_interfaces(svc_ref, intfs):
        # type: (ServiceReference, List[str]) -> bool
        if not intfs:
            return False

        object_class = svc_ref.get_property(constants.OBJECTCLASS)
        for item in intfs:
            if not item in object_class:
                return False

        return True

    def _find_existing_export_endpoint(self, svc_ref, cid):
        # type: (ServiceReference, str) -> Optional[ExportRegistration]
        for er in self._exported_regs:
            if er.match_sr(svc_ref, cid):
                return er
        return None

    def _add_exported_service(self, export_reg):
        # type: (ExportRegistration) -> None
        with self._exported_regs_lock:
            self._exported_regs.append(export_reg)

    def _remove_exported_service(self, export_reg):
        # type: (ExportRegistration) -> None
        with self._exported_regs_lock:
            self._exported_regs.remove(export_reg)

    def _add_imported_service(self, import_reg):
        # type: (ImportRegistration) -> None
        with self._imported_regs_lock:
            self._imported_regs.append(import_reg)

    def _remove_imported_service(self, import_reg):
        # type: (ImportRegistration) -> None
        with self._imported_regs_lock:
            self._imported_regs.remove(import_reg)


# ------------------------------------------------------------------------------
# Internal class used to implement ExportRegistration/ExportReference below.
class _ExportEndpoint(object):
    def __init__(self, rsa, export_container, ed, svc_ref):
        # type: (RemoteServiceAdminImpl, ExportContainer, EndpointDescription, ServiceReference) -> None
        assert rsa
        self.__rsa = rsa
        assert export_container
        self.__export_container = export_container
        assert ed
        self.__ed = ed
        assert svc_ref
        self.__svc_ref = svc_ref
        self.__lock = threading.RLock()
        self.__active_registrations = []  # type: List[ExportRegistration]
        self.__orig_props = self.__ed.get_properties()

    def _rsa(self):
        with self.__lock:
            return self.__rsa

    def _originalprops(self):
        # type: () -> Dict[str, Any]
        with self.__lock:
            return self.get_reference().get_properties()

    def _add_export_registration(self, export_reg):
        # type: (ExportRegistration) -> None
        with self.__lock:
            self.__active_registrations.append(export_reg)

    def _remove_export_registration(self, export_reg):
        # type: (ExportRegistration) -> None
        with self.__lock:
            self.__active_registrations.remove(export_reg)

    def get_description(self):
        # type: () -> EndpointDescription
        with self.__lock:
            return self.__ed

    def get_reference(self):
        # type: () -> ServiceReference
        with self.__lock:
            return self.__svc_ref

    def get_export_container_id(self):
        # type: () -> str
        with self.__lock:
            return self.__export_container.get_id()

    def get_remoteservice_id(self):
        # type: () -> Tuple[Tuple[str, str], int]
        with self.__lock:
            return self.__ed.get_remoteservice_id()

    def update(self, props):
        # type: (Dict[str, Any]) -> EndpointDescription
        with self.__lock:
            rsprops = self.__orig_props.copy()
            if not props:
                updatedprops = rsprops
            else:
                updatedprops = props.copy()
                updatedprops.update(rsprops)

            updatedprops.update(self.__svc_ref.get_properties())
            updatedprops[ECF_ENDPOINT_TIMESTAMP] = get_current_time_millis()
            self.__ed = EndpointDescription(self.__svc_ref, updatedprops)
            return self.__ed

    def close(self, export_reg):
        # type: (ExportRegistration) -> bool
        with self.__lock:
            try:
                self.__active_registrations.remove(export_reg)
            except ValueError:
                pass
            if len(self.__active_registrations) == 0:
                try:
                    self.__export_container.unexport_service(self.__ed)
                except:
                    _logger.exception(
                        "get_exception in exporter.unexport_service ed=%s",
                        self.__ed,
                    )

                # pylint: disable=W0212
                self.__rsa._remove_exported_service(export_reg)
                self.__ed = (
                    self.__export_container
                ) = self.__svc_ref = self.__rsa = None
                return True
        return False


# ------------------------------------------------------------------------------


class ExportReferenceImpl(ExportReference):
    """
    Implementation of ExportReference API. See ExportReference class for
    external contract and documentation
    """

    @classmethod
    def fromendpoint(cls, endpoint):
        # type: (_ExportEndpoint) -> ExportReference
        return cls(endpoint=endpoint)

    @classmethod
    def fromexception(cls, e, ed):
        # type: (Optional[Tuple[Any, Any, Any]], EndpointDescription) -> ExportReference
        return cls(endpoint=None, exception=e, errored=ed)

    def __init__(self, endpoint=None, exception=None, errored=None):
        # type: (Optional[_ExportEndpoint], Optional[Tuple[Any, Any, Any]], Optional[EndpointDescription]) -> None
        self.__lock = threading.RLock()
        if endpoint is None:
            if exception is None or errored is None:
                raise RemoteServiceError(
                    "Must supply either endpoint or "
                    "throwable/error EndpointDescription"
                )
            self.__exception = exception
            self.__errored = errored
            self._endpoint = None
        else:
            self._endpoint = endpoint
            self.__exception = self.__errored = None

    def get_export_container_id(self):
        # type: () -> Optional[Tuple[str, str]]
        with self.__lock:
            return (
                None
                if self._endpoint is None
                else self._endpoint.get_export_container_id()
            )

    def get_remoteservice_id(self):
        # type: () -> Optional[Tuple[str, str]]
        with self.__lock:
            return (
                None
                if self._endpoint is None
                else self._endpoint.get_remoteservice_id()
            )

    def get_reference(self):
        # type: () -> Optional[ServiceReference]
        with self.__lock:
            return (
                None
                if self._endpoint is None
                else self._endpoint.get_reference()
            )

    def get_description(self):
        # type: () -> Optional[EndpointDescription]
        with self.__lock:
            return (
                self.__errored
                if self._endpoint is None
                else self._endpoint.get_description()
            )

    def get_exception(self):
        # type: () -> Tuple[Any, Any, Any]
        with self.__lock:
            return self.__exception

    def update(self, properties):
        # type: (Dict[str, Any]) -> EndpointDescription
        with self.__lock:
            return (
                None
                if self._endpoint is None
                else self._endpoint.update(properties)
            )

    def close(self, export_reg):
        # type: (ExportRegistration) -> bool
        with self.__lock:
            if self._endpoint is None:
                return False

            result = self._endpoint.close(export_reg)
            self._endpoint = None
            return bool(result)


# ------------------------------------------------------------------------------


class ExportRegistrationImpl(ExportRegistration):
    """
    Implementation of ExportRegistration API.
    See ExportRegistration class for external contract and documentation
    """

    @classmethod
    def fromreg(cls, export_reg):
        # type: (ExportRegistrationImpl) -> ExportRegistrationImpl
        # pylint: disable=W0212
        return cls(export_reg.__rsa, export_reg.__exportref._endpoint)

    @classmethod
    def fromendpoint(cls, rsa, exporter, ed, svc_ref):
        # type: (RemoteServiceAdminImpl, ExportContainer, EndpointDescription, ServiceReference) -> ExportRegistration
        return cls(rsa, _ExportEndpoint(rsa, exporter, ed, svc_ref))

    @classmethod
    def fromexception(cls, e, ed):
        # type: (Tuple[Any, Any, Any], EndpointDescription) -> ExportRegistration
        return cls(rsa=None, endpoint=None, exception=e, errored=ed)

    def __init__(self, rsa=None, endpoint=None, exception=None, errored=None):
        # type: (Optional[RemoteServiceAdminImpl], Optional[_ExportEndpoint], Optional[Tuple[Any, Any, Any]], Optional[EndpointDescription]) -> None
        # pylint: disable=W0212
        if endpoint is None:
            if exception is None or errored is None:
                raise RemoteServiceError(
                    "export endpoint or get_exception/errorED must not be null"
                )
            self.__exportref = ExportReferenceImpl.fromexception(
                exception, errored
            )  # type: ExportReferenceImpl
            self.__rsa = None
        else:
            self.__rsa = endpoint._rsa()
            endpoint._add_export_registration(self)
            self.__exportref = ExportReferenceImpl.fromendpoint(
                endpoint
            )  # type: ExportReferenceImpl

        self.__closed = False
        self.__updateexception = None
        self.__lock = threading.RLock()

    def match_sr(self, svc_ref, cid=None):
        # type: (ServiceReference, Optional[Tuple[str, str]] ) -> bool
        """
        Checks if this export registration matches the given service reference

        :param svc_ref: A service reference
        :param cid: A container ID
        :return: True if the service matches this export registration
        """
        with self.__lock:
            our_sr = self.get_reference()
            if our_sr is None:
                return False

            sr_compare = our_sr == svc_ref
            if cid is None:
                return sr_compare

            our_cid = self.get_export_container_id()
            if our_cid is None:
                return False

            return sr_compare and our_cid == cid

    def get_export_reference(self):
        # type: () -> ExportReference
        """
        Returns the reference matching this registration

        :return: An export reference
        """
        with self.__lock:
            return None if self.__closed else self.__exportref

    def _exportendpoint(self, svc_ref, cid):
        # type: (ServiceReference, Tuple[str, str]) -> Optional[_ExportEndpoint]
        with self.__lock:
            # pylint: disable=W0212
            return (
                None
                if self.__closed
                else self.__exportref._endpoint
                if self.match_sr(svc_ref, cid)
                else None
            )

    def get_export_container_id(self):
        # type: () -> Optional[Tuple[str, str]]
        """
        Returns the export container ID

        :return: An export container ID
        """
        with self.__lock:
            return (
                None
                if self.__closed
                else self.__exportref.get_export_container_id()
            )

    def get_remoteservice_id(self):
        # type: () -> Optional[Tuple[Tuple[str, str], int]]
        """
        Returns the remote service ID

        :return: The remote service ID
        """
        with self.__lock:
            return (
                None
                if self.__closed
                else self.__exportref.get_remoteservice_id()
            )

    def get_reference(self):
        # type: () -> Optional[ServiceReference]
        """
        Retruns the service reference of the exported service

        :return: A service reference
        """
        with self.__lock:
            return None if self.__closed else self.__exportref.get_reference()

    def get_exception(self):
        # type: () -> Optional[Tuple[Any, Any, Any]]
        """
        Returns the exception associated to the export

        :return: An exception tuple, if any
        """
        with self.__lock:
            return (
                self.__updateexception
                if self.__updateexception or self.__closed
                else self.__exportref.get_exception()
            )

    def get_description(self):
        # type: () -> Optional[EndpointDescription]
        """
        Returns the description of the endpoint

        :return: An endpoint description
        """
        with self.__lock:
            return None if self.__closed else self.__exportref.get_description()

    def update(self, properties):
        # type: (Dict[str, Any]) -> Optional[EndpointDescription]
        with self.__lock:
            if self.__closed:
                self.__updateexception = ValueError(
                    "Update failed since ExportRegistration already closed"
                )
                return None
            # if properties is set then copy
            props = properties.copy() if properties else dict()
            try:
                updated_ed = self.__exportref.update(props)
            except Exception as e:
                self.__updateexception = e
                return None

            if not updated_ed:
                self.__updateexception = ValueError(
                    "Update failed because ExportEndpoint was None"
                )
                return None

            # pylint: disable=W0212
            self.__updateexception = None
            if self.__rsa:
                self.__rsa._publish_event(
                    RemoteServiceAdminEvent.fromexportupdate(
                        self.__rsa._get_bundle(), self
                    )
                )
            return updated_ed

    def close(self):
        """
        Cleans up the export endpoint
        """
        publish = False
        exporterid = rsid = exception = export_ref = ed = None
        with self.__lock:
            if not self.__closed:
                exporterid = self.__exportref.get_export_container_id()
                export_ref = self.__exportref
                rsid = self.__exportref.get_remoteservice_id()
                ed = self.__exportref.get_description()
                exception = self.__exportref.get_exception()
                self.__closed = True
                publish = self.__exportref.close(self)
                self.__exportref = None

        # pylint: disable=W0212
        if publish and export_ref and self.__rsa:
            self.__rsa._publish_event(
                RemoteServiceAdminEvent.fromexportunreg(
                    self.__rsa._get_bundle(),
                    exporterid,
                    rsid,
                    export_ref,
                    exception,
                    ed,
                )
            )
            self.__rsa = None


class _ImportEndpoint(object):
    def __init__(self, rsa, import_container, ed, svc_reg):
        # type: (RemoteServiceAdminImpl, ImportContainer, EndpointDescription, ServiceRegistration) -> None
        assert rsa
        self.__rsa = rsa
        assert import_container
        self.__importer = import_container
        assert ed
        self.__ed = ed
        assert svc_reg
        self.__svc_reg = svc_reg
        self.__lock = threading.RLock()
        self.__active_registrations = []  # type: List[ImportRegistration]

    def _add_import_registration(self, import_reg):
        # type: (ImportRegistration) -> None
        with self.__lock:
            self.__active_registrations.append(import_reg)

    def _rsa(self):
        # type: () -> RemoteServiceAdminImpl
        return self.__rsa

    def match_ed(self, ed):
        # type: (EndpointDescription) -> bool
        with self.__lock:
            if len(self.__active_registrations) == 0:
                return False
            return self.__ed.is_same_service(ed)

    def get_reference(self):
        # type: () -> ServiceReference
        with self.__lock:
            return (
                None
                if self.__importer is None
                else self.__svc_reg.get_reference()
            )

    def get_description(self):
        # type: () -> EndpointDescription
        with self.__lock:
            return self.__ed

    def get_import_container_id(self):
        # type: () -> str
        with self.__lock:
            return None if self.__importer is None else self.__importer.get_id()

    def get_export_container_id(self):
        # type: () -> Tuple[str, str]
        with self.__lock:
            return self.__ed.get_container_id()

    def get_remoteservice_id(self):
        # type: () -> Tuple[Tuple[str, str], int]
        with self.__lock:
            return self.__ed.get_remoteservice_id()

    def update(self, ed):
        # type: (EndpointDescription) -> Optional[EndpointDescription]
        with self.__lock:
            if self.__svc_reg is None:
                return None

            # Prepare properties
            # pylint: disable=W0212
            ed_props = self.__ed.get_properties()
            new_props = self.__importer._prepare_proxy_props(ed)
            ed_props.update(new_props)

            # Make a new endpoint description
            self.__ed = EndpointDescription.fromprops(ed_props)

            # Update the exported service
            self.__svc_reg.set_properties(self.__ed.get_properties())
            return self.__ed

    def close(self, import_reg):
        with self.__lock:
            try:
                self.__active_registrations.remove(import_reg)
            except ValueError:
                pass

            if not self.__active_registrations:
                if self.__svc_reg is not None:
                    try:
                        self.__svc_reg.unregister()
                    except BundleException:
                        # The service might already have unregistered
                        pass
                    except:
                        _logger.exception(
                            "Exception unregistering local proxy=%s",
                            self.__svc_reg.get_reference(),
                        )
                    self.__svc_reg = None
                try:
                    self.__importer.unimport_service(self.__ed)
                except:
                    _logger.exception(
                        "Exception calling importer.unimport_service with ed=%s",
                        self.__ed,
                    )
                    return False

                # pylint: disable=W0212
                self.__rsa._remove_imported_service(import_reg)
                self.__importer = self.__ed = self.__rsa = None
                return True

        return False


# ------------------------------------------------------------------------------
# Implementation of ExportReference API.  See ExportReference class for external
# contract and documentation
class ImportReferenceImpl(ImportReference):
    @classmethod
    def fromendpoint(cls, endpoint):
        # type: (_ImportEndpoint) -> ImportReferenceImpl
        return cls(endpoint=endpoint)

    @classmethod
    def fromexception(cls, e, errored):
        # type: (Tuple[Any, Any, Any], EndpointDescription) -> ImportReferenceImpl
        return cls(endpoint=None, exception=e, errored=errored)

    def __init__(self, endpoint=None, exception=None, errored=None):
        # type: (Optional[_ImportEndpoint], Optional[Tuple[Any, Any, Any]], Optional[EndpointDescription]) -> None
        self.__lock = threading.RLock()
        if endpoint is None:
            if exception is None or errored is None:
                raise RemoteServiceError(
                    "Must supply either endpoint or "
                    "throwable/errorEndpointDescription"
                )
            self.__exception = exception
            self.__errored = errored
            self.__endpoint = None
        else:
            self.__endpoint = endpoint
            self.__exception = self.__errored = None

    def _importendpoint(self):
        # type: () -> _ImportEndpoint
        with self.__lock:
            return self.__endpoint

    def match_ed(self, ed):
        # type: (EndpointDescription) -> bool
        with self.__lock:
            return (
                None
                if self.__endpoint is None
                else self.__endpoint.match_ed(ed)
            )

    def get_import_container_id(self):
        # type: () -> str
        with self.__lock:
            return (
                None
                if self.__endpoint is None
                else self.__endpoint.get_import_container_id()
            )

    def get_export_container_id(self):
        with self.__lock:
            return (
                None
                if self.__endpoint is None
                else self.__endpoint.get_export_container_id()
            )

    def get_remoteservice_id(self):
        # type: () -> Optional[Tuple[Tuple[str, str], int]]
        with self.__lock:
            return (
                None
                if self.__endpoint is None
                else self.__endpoint.get_remoteservice_id()
            )

    def get_reference(self):
        # type: () -> Optional[ServiceReference]
        with self.__lock:
            return (
                None
                if self.__endpoint is None
                else self.__endpoint.get_reference()
            )

    def get_description(self):
        # type: () -> Optional[EndpointDescription]
        with self.__lock:
            return (
                self.__errored
                if self.__endpoint is None
                else self.__endpoint.get_description()
            )

    def get_exception(self):
        # type: () -> Optional[Tuple[Any, Any, Any]]
        with self.__lock:
            return self.__exception

    def update(self, endpoint):
        # type: (EndpointDescription) -> Optional[EndpointDescription]
        with self.__lock:
            return (
                None
                if self.__endpoint is None
                else self.__endpoint.update(endpoint)
            )

    def close(self, import_reg):
        with self.__lock:
            if self.__endpoint is None:
                return False

            result = self.__endpoint.close(import_reg)
            self.__endpoint = None
            return result


# ------------------------------------------------------------------------------


class ImportRegistrationImpl(ImportRegistration):
    """
    Implementation of ExportRegistration API.

    See ExportRegistration class for external contract and documentation
    """

    @classmethod
    def fromendpoint(cls, rsa, importer, ed, svc_reg):
        # type: (RemoteServiceAdminImpl, ImportContainer, EndpointDescription, ServiceRegistration) -> ImportRegistration
        return cls(endpoint=_ImportEndpoint(rsa, importer, ed, svc_reg))

    @classmethod
    def fromexception(cls, e, ed):
        # type: (Tuple[Any, Any, Any], EndpointDescription) -> ImportRegistration
        return cls(endpoint=None, exception=e, errored=ed)

    @classmethod
    def fromreg(cls, reg):
        # type: (ImportRegistrationImpl) -> ImportRegistrationImpl
        # pylint: disable=W0212
        return cls(endpoint=reg.__importref._importendpoint())

    def __init__(self, endpoint=None, exception=None, errored=None):
        # type: (Optional[_ImportEndpoint], Optional[Tuple[Any, Any, Any]], Optional[EndpointDescription]) -> None
        # pylint: disable=W0212
        if endpoint is None:
            if exception is None or errored is None:
                raise RemoteServiceError(
                    "export endpoint or get_exception/errorED must not be null"
                )
            self.__importref = ImportReferenceImpl.fromexception(
                exception, errored
            )
            self.__rsa = None
        else:
            self.__rsa = endpoint._rsa()
            endpoint._add_import_registration(self)
            self.__importref = ImportReferenceImpl.fromendpoint(endpoint)

        self.__closed = False
        self.__updateexception = None
        self.__lock = threading.RLock()

    def _import_endpoint(self):
        # type: () -> _ImportEndpoint
        # pylint: disable=W0212
        with self.__lock:
            return None if self.__closed else self.__importref._importendpoint()

    def match_ed(self, ed):
        # type: (EndpointDescription) -> bool
        with self.__lock:
            return False if self.__closed else self.__importref.match_ed(ed)

    def get_import_reference(self):
        # type: () -> ImportReference
        with self.__lock:
            return None if self.__closed else self.__importref

    def get_import_container_id(self):
        # type: () -> str
        with self.__lock:
            return (
                None
                if self.__closed
                else self.__importref.get_import_container_id()
            )

    def get_export_container_id(self):
        with self.__lock:
            return (
                None
                if self.__closed
                else self.__importref.get_export_container_id()
            )

    def get_remoteservice_id(self):
        # type: () -> Optional[Tuple[Tuple[str, str], int]]
        with self.__lock:
            return (
                None
                if self.__closed is None
                else self.__importref.get_remoteservice_id()
            )

    def get_reference(self):
        # type: () -> Optional[ServiceReference]
        with self.__lock:
            return None if self.__closed else self.__importref.get_reference()

    def get_exception(self):
        # type: () -> Optional[Tuple[Any, Any, Any]]
        with self.__lock:
            return (
                self.__updateexception
                if self.__updateexception or self.__closed
                else self.__importref.get_exception()
            )

    def get_description(self):
        # type: () -> Optional[EndpointDescription]
        with self.__lock:
            return None if self.__closed else self.__importref.get_description()

    def update(self, endpoint_description):
        # type: (EndpointDescription) -> bool
        with self.__lock:
            if self.__closed:
                self.__updateexception = ValueError(
                    "Update failed since ImportRegistration already closed"
                )
                return False

            try:
                self.__importref.update(endpoint_description)
            except Exception as e:
                self.__updateexception = e
                return False

            if self.__rsa:
                # pylint: disable=W0212
                self.__rsa._publish_event(
                    RemoteServiceAdminEvent.fromimportupdate(
                        self.__rsa._get_bundle(), self
                    )
                )
                return True

            return False

    def close(self):
        publish = False
        importerid = rsid = import_ref = exception = ed = None
        with self.__lock:
            if not self.__closed:
                importerid = self.__importref.get_import_container_id()
                rsid = self.__importref.get_remoteservice_id()
                import_ref = self.__importref
                exception = self.__importref.get_exception()
                ed = self.__importref.get_description()
                self.__closed = True
                publish = self.__importref.close(self)
                self.__importref = None
        if publish and import_ref and self.__rsa:
            # pylint: disable=W0212
            self.__rsa._publish_event(
                RemoteServiceAdminEvent.fromimportunreg(
                    self.__rsa._get_bundle(),
                    importerid,
                    rsid,
                    import_ref,
                    exception,
                    ed,
                )
            )
            self.__rsa = None


# ------------------------------------------------------------------------------


class DebugRemoteServiceAdminListener(RemoteServiceAdminListener):
    """
    Implementation of RemoteServiceAdminListener that supports debugging by
    printing out information about the RemoteServiceAdminEvents.
    """

    EXPORT_MASK = (
        RemoteServiceAdminEvent.EXPORT_ERROR
        | RemoteServiceAdminEvent.EXPORT_REGISTRATION
        | RemoteServiceAdminEvent.EXPORT_UNREGISTRATION
        | RemoteServiceAdminEvent.EXPORT_WARNING
    )

    IMPORT_MASK = (
        RemoteServiceAdminEvent.IMPORT_ERROR
        | RemoteServiceAdminEvent.IMPORT_REGISTRATION
        | RemoteServiceAdminEvent.IMPORT_UNREGISTRATION
        | RemoteServiceAdminEvent.IMPORT_WARNING
    )

    ALL_MASK = EXPORT_MASK | IMPORT_MASK

    def __init__(
        self,
        file=sys.stdout,
        event_mask=ALL_MASK,
        write_endpoint=True,
        ed_encoding="unicode",
        xml_declaration=True,
    ):
        self._output = file
        self._writer = EDEFWriter(ed_encoding, xml_declaration)
        self._event_mask = event_mask
        self._write_endpoint = write_endpoint
        self._eventtypestr = {
            RemoteServiceAdminEvent.EXPORT_ERROR: "EXPORT_ERROR",
            RemoteServiceAdminEvent.EXPORT_REGISTRATION: "EXPORT_REGISTRATION",
            RemoteServiceAdminEvent.EXPORT_UNREGISTRATION: "EXPORT_UNREGISTRATION",
            RemoteServiceAdminEvent.EXPORT_UPDATE: "EXPORT_UPDATE",
            RemoteServiceAdminEvent.EXPORT_WARNING: "EXPORT_WARNING",
            RemoteServiceAdminEvent.IMPORT_ERROR: "IMPORT_ERROR",
            RemoteServiceAdminEvent.IMPORT_REGISTRATION: "IMPORT_REGISTRATION",
            RemoteServiceAdminEvent.IMPORT_UNREGISTRATION: "IMPORT_UNREGISTRATION",
            RemoteServiceAdminEvent.IMPORT_UPDATE: "IMPORT_UPDATE",
            RemoteServiceAdminEvent.IMPORT_WARNING: "IMPORT_WARNING",
        }
        self._exporttypes = [
            RemoteServiceAdminEvent.EXPORT_REGISTRATION,
            RemoteServiceAdminEvent.EXPORT_UNREGISTRATION,
            RemoteServiceAdminEvent.EXPORT_UPDATE,
            RemoteServiceAdminEvent.EXPORT_WARNING,
        ]
        self._importtypes = [
            RemoteServiceAdminEvent.IMPORT_REGISTRATION,
            RemoteServiceAdminEvent.IMPORT_UNREGISTRATION,
            RemoteServiceAdminEvent.IMPORT_UPDATE,
            RemoteServiceAdminEvent.IMPORT_WARNING,
        ]
        self._errortypes = [
            RemoteServiceAdminEvent.EXPORT_ERROR,
            RemoteServiceAdminEvent.IMPORT_ERROR,
        ]

    def write_description(self, ed):
        if self._write_endpoint and ed:
            self._output.write("---Endpoint Description---\n")
            self._output.write(self._writer.to_string([ed]))
            self._output.write("\n---End Endpoint Description---\n")
            self._output.flush()

    def write_ref(self, svc_ref, cid, rsid, ed):
        if svc_ref:
            self._output.write(str(svc_ref) + ";")
        self._output.write("local=" + str(cid))
        # get_remoteservice_id should be of form:  ((ns,cid,get_remoteservice_id)
        self._output.write(";remote=")
        if isinstance(rsid, tuple) and len(list(rsid)) == 2:
            nscid = rsid[0]
            if isinstance(nscid, tuple) and len(list(nscid)) == 2:
                self._output.write(str(nscid[1]) + ":" + str(rsid[1]))
        else:
            self._output.write(str(rsid))
        self._output.write("\n")
        self._output.flush()
        self.write_description(ed)

    def write_exception(self, exception):
        self._output.write("---Exception Stack---\n")
        print_exception(
            exception[0],
            exception[1],
            exception[2],
            limit=None,
            file=self._output,
        )
        self._output.write("---End Exception Stack---\n")

    def write_type(self, event_type):
        (dt, micro) = datetime.now().strftime("%H:%M:%S.%f").split(".")
        dt = "%s.%03d" % (dt, int(micro) / 1000)
        self._output.write(
            dt + ";" + self._eventtypestr.get(event_type, "UNKNOWN") + ";"
        )

    def write_event(self, rsa_event):
        event_type = rsa_event.get_type()
        rs_ref = None
        svc_ref = None
        exception = None
        self.write_type(event_type)
        if event_type in self._exporttypes:
            rs_ref = rsa_event.get_export_ref()
        elif event_type in self._importtypes:
            rs_ref = rsa_event.get_import_ref()
        elif event_type in self._errortypes:
            exception = rsa_event.get_exception()
        if rs_ref:
            svc_ref = rs_ref.get_reference()
        self.write_ref(
            svc_ref,
            rsa_event.get_container_id(),
            rsa_event.get_remoteservice_id(),
            rsa_event.get_description(),
        )
        if exception:
            self.write_exception(exception)

    def remote_admin_event(self, rsa_event):
        self.write_event(rsa_event)



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
iPOPO handlers constants and base classes

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

SERVICE_IPOPO_HANDLER_FACTORY = "ipopo.handler.factory"
"""
iPOPO handler factory service specification. Those services should implement
the methods of HandlerFactory.
"""

PROP_HANDLER_ID = "ipopo.handler.id"
""" Service property: the ID of the iPOPO handler factory """

# ------------------------------------------------------------------------------

KIND_PROPERTIES = "properties"
"""
Represents the 'properties' kind of handler, which manipulates the component
to notify property changes.
"""

KIND_DEPENDENCY = "dependency"
"""
Represents the 'dependency' kind of handler.
Those handlers must implement the following methods:

* get_bindings(): Retrieves the list of bound service references
* is_valid(): Returns True if the dependency is in a valid state
"""

KIND_SERVICE_PROVIDER = "service_provider"
"""
Represents the 'service_provider' kind of handler.
Those handlers must implement the following method:

* get_service_reference(): Retrieves the reference of the provided service
  (a ServiceReference object).

It should also implement the following ones:

* on_controller_changer(): Called when a component controller has been
  modified. The publication of a service might be stopped if its controller is
  set to False.
* on_property_change(): Called when a component property has been modified.
  The provided service properties should be modified accordingly.
"""

# ------------------------------------------------------------------------------


class HandlerFactory(object):
    # pylint: disable=R0903
    """
    Handler factory abstract class
    """

    def get_handlers(self, component_context, instance):
        """
        Prepares handlers for the given component

        :param component_context: The ComponentContext bean
        :param instance: The component instance
        :return: The list of handlers associated to the given component
        """
        pass


# ------------------------------------------------------------------------------


class Handler(object):
    """
    Basic handler abstract class
    """

    def get_kinds(self):
        # pylint: disable=R0201
        """
        Returns the kinds of this handler

        :return: A tuple of the kinds of this handler, or None
        """
        pass

    def manipulate(self, stored_instance, component_instance):
        """
        Manipulates the associated component instance
        """
        pass

    def check_event(self, event):
        # pylint: disable=R0201, W0613
        """
        Tests if the given service event must be handled or ignored, based
        on the state of the iPOPO service and on the content of the event.

        :param event: A service event
        :return: True if the event can be handled, False if it must be ignored
        """
        return True

    def is_valid(self):
        # pylint: disable=R0201
        """
        Checks this handler is valid. All handlers must be valid for a
        component to be validated

        :return: True if the handler is in a valid state
        """
        return True

    def on_controller_change(self, name, value):
        """
        Notifies the change of state of the controller with the given name

        :param name: The name of the controller
        :param value: The new value of the controller
        """
        pass

    def on_property_change(self, name, old_value, new_value):
        """
        Handles a property changed event

        :param name: The changed property name
        :param old_value: The previous property value
        :param new_value: The new property value
        """
        pass

    def start(self):
        """
        Starts the handler (listeners, ...). Called once, after the component
        has been manipulated by all handlers.
        """
        pass

    def stop(self):
        """
        Stops the handler. Called once, just after the component has been
        killed
        """
        pass

    def clear(self):
        """
        Called just after a component has been killed and all handlers have
        been stopped. The handler should release all its resources here.
        """
        pass

    def pre_validate(self):
        """
        Called just before a component is validated
        """
        pass

    def post_validate(self):
        """
        Called just after a component has been validated
        """
        pass

    def pre_invalidate(self):
        """
        Called just before a component is invalidated
        """
        pass

    def post_invalidate(self):
        """
        Called just after a component has been invalidated
        """
        pass


class HandlerException(Exception):
    """
    Kind of exception used by handlers
    """

    pass


# ------------------------------------------------------------------------------


class ServiceProviderHandler(Handler):
    """
    Service provider handler abstract class
    """

    def get_service_reference(self):
        # pylint: disable=R0201
        """
        Returns the reference to the service provided by this handler
        """
        return None


# ------------------------------------------------------------------------------


class DependencyHandler(Handler):
    """
    Dependency handler abstract class
    """

    def get_field(self):
        # pylint: disable=R0201
        """
        Returns the name of the field where to inject the dependency
        """
        return None

    def try_binding(self):
        # pylint: disable=R0201
        """
        Forces the handler to try to bind to existing services
        """
        pass

    def get_bindings(self):
        # pylint: disable=R0201
        """
        Retrieves the list of the references to the bound services

        :return: A list of ServiceReferences objects
        """
        return None

    def get_value(self):
        # pylint: disable=R0201
        """
        Returns the value to inject
        """
        return None



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix remote services: MQTT pseudo RPC

An remote service protocol based on MQTT.
This implementation tries to mimic the MQTT-RPC Javascript project:
https://github.com/wolfeidau/mqtt-rpc

This module depends on the paho-mqtt package (ex-mosquitto), provided by the
Eclipse Foundation: see http://www.eclipse.org/paho

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import json
import logging
import threading
import uuid

# MQTT client
import pelix.misc.mqtt_client

# iPOPO decorators
from pelix.ipopo.decorators import (
    ComponentFactory,
    Property,
    Provides,
    Validate,
    Invalidate,
)

# Pelix & Remote services
from pelix.utilities import to_str
from pelix.remote import RemoteServiceError
import pelix.remote
import pelix.remote.transport.commons as commons

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

MQTTRPC_CONFIGURATION = "mqttrpc"
""" Remote Service configuration constant """

PROP_MQTT_TOPIC = "mqttrpc.topic"
""" Remote Service property: topic to use to contact the service """

_logger = logging.getLogger(__name__)

# Topic suffixes
TOPIC_REQUEST = "request"
TOPIC_RESPONSE = "reply"

# JSON dictionary keys
KEY_ERROR = "err"
KEY_DATA = "data"
KEY_CORRELATION_ID = "_correlationId"
KEY_SENDER = pelix.remote.PROP_ENDPOINT_FRAMEWORK_UUID

# ------------------------------------------------------------------------------


def make_topic(topic, suffix):
    # pylint: disable=W0613
    """
    Prepares a topic with the given suffix

    :param topic: MQTT topic
    :param suffix: String to append to the topic
    :return: The suffixed topic
    """
    return "{0}/{1}".format(topic, TOPIC_REQUEST).replace("//", "/")


@ComponentFactory(pelix.remote.FACTORY_TRANSPORT_MQTTRPC_EXPORTER)
@Provides(pelix.remote.SERVICE_EXPORT_PROVIDER)
@Property("_topic", PROP_MQTT_TOPIC, "pelix/remote-services/mqttrpc/{fw_uid}")
@Property(
    "_kinds",
    pelix.remote.PROP_REMOTE_CONFIGS_SUPPORTED,
    (MQTTRPC_CONFIGURATION,),
)
@Property("_host", "mqtt.host", "localhost")
@Property("_port", "mqtt.port", 1883)
class MqttRpcServiceExporter(commons.AbstractRpcServiceExporter):
    """
    MQTT-RPC Remote Services exporter
    """

    def __init__(self):
        """
        Sets up the exporter
        """
        # Call parent
        super(MqttRpcServiceExporter, self).__init__()

        # Handled configurations
        self._kinds = None

        # MQTT topic
        self._topic = ""
        self.__real_topic = ""

        # Topic to reply to requests
        self.__reply_topic = ""

        # MQTT server
        self._host = None
        self._port = None

        # MQTT client
        self.__mqtt = None

    @Validate
    def validate(self, context):
        """
        Component validated
        """
        # Call the parent
        super(MqttRpcServiceExporter, self).validate(context)

        # Format the topic prefix
        self.__real_topic = self._topic.format(fw_uid=self._framework_uid)

        # Normalize topic
        self.__reply_topic = make_topic(self.__real_topic, TOPIC_RESPONSE)

        # Create the MQTT client
        self.__mqtt = pelix.misc.mqtt_client.MqttClient()

        # Customize callbacks
        self.__mqtt.on_connect = self.__on_connect
        self.__mqtt.on_message = self.__on_message

        # Prepare the connection
        self.__mqtt.connect(self._host, self._port)

    @Invalidate
    def invalidate(self, context):
        """
        Component invalidated
        """
        # Disconnect from the server (this stops the loop)
        self.__mqtt.disconnect()

        # Call the parent
        super(MqttRpcServiceExporter, self).invalidate(context)

        # Clean up members
        self.__mqtt = None

    def __on_connect(self, client, result_code):
        # pylint: disable=W0613
        """
        Client connected to the server
        """
        if not result_code:
            # Connection is OK, subscribe to the topic
            # Subscribe to endpoints calls
            request_topic_filter = make_topic(self.__real_topic, TOPIC_REQUEST)
            self.__mqtt.subscribe(request_topic_filter)

    def __on_message(self, client, msg):
        # pylint: disable=W0613
        """
        An MQTT message has been received

        :param client: MQTT client
        :param msg: A MQTTMessage bean
        """
        try:
            # Parse the message
            data = json.loads(to_str(msg.payload))
        except ValueError as ex:
            # Bad content
            _logger.error("Error reading MQTT-RPC request: %s", ex)
            return

        try:
            if data[KEY_SENDER] == self._framework_uid:
                # We published this message
                return
        except KeyError:
            # Not sent by us
            pass

        # Handle the request in a different thread
        threading.Thread(
            name="MQTT-RPC-Exporter", target=self.__handle_rpc, args=(data,)
        ).start()

    def __handle_rpc(self, data):
        """
        Handles an RPC request (should be called in a specific thread)

        :param data: RPC description
        """
        # Prepare the result dictionary
        result = {
            KEY_ERROR: None,
            KEY_DATA: None,
            KEY_CORRELATION_ID: data.get(KEY_CORRELATION_ID),
            KEY_SENDER: self._framework_uid,
        }

        # Extract parameters
        try:
            call_info = data[KEY_DATA]
            if not call_info:
                raise ValueError("No call information given")

            method = call_info[0]
            params = call_info[1:]
        except (KeyError, IndexError, ValueError) as ex:
            result[KEY_ERROR] = "Endpoint name is missing: {0}".format(ex)
        else:
            try:
                # Call the service
                result[KEY_DATA] = self.dispatch(method, params)
            except Exception as ex:
                # An error occurred
                result[KEY_ERROR] = str(ex)

        try:
            # Publish the result
            self.__mqtt.publish(self.__reply_topic, json.dumps(result), qos=2)
        except (ValueError, AttributeError) as ex:
            _logger.error("Error replying an RPC request: %s", ex)

    def make_endpoint_properties(self, svc_ref, name, fw_uid):
        """
        Prepare properties for the ExportEndpoint to be created

        :param svc_ref: Service reference
        :param name: Endpoint name
        :param fw_uid: Framework UID
        :return: A dictionary of extra endpoint properties
        """
        return {PROP_MQTT_TOPIC: self.__real_topic}


# ------------------------------------------------------------------------------


class _MqttCallableProxy(object):
    # pylint: disable=R0903
    """
    Callable object that makes the real request to the MQTT server
    """

    def __init__(self, uid, topic, method, publish_method):
        """
        Stores parameters

        :param uid: Endpoint UID
        :param topic: Endpoint topic
        :param method: Name of the method to call
        :param publish_method: Method to call to publish the request message
        """
        # Local information
        self.__uid = uid
        self.__topic = topic
        self.__method_name = method
        self.__publish = publish_method

        # Event to wait for the result
        self._event = threading.Event()

        # Result
        self._error = None
        self._result = None

    def handle_result(self, result, error):
        """
        The result has been received

        :param result: Call result
        :param error: Error message
        """
        if not self._error and not self._result:
            # Store results, if not already set
            self._error = error
            self._result = result

        # Unlock the call
        self._event.set()

    def __call__(self, *args, **kwargs):
        """
        Method call
        """
        # Send a request
        request = [self.__method_name]
        if args:
            request.extend(args)

        # Keyword arguments are ignores
        self.__publish(self.__uid, self, self.__topic, request)

        # Wait for an answer
        self._event.wait()

        # Act accordingly
        if self._error:
            raise RemoteServiceError(self._error)

        else:
            return self._result


class _ServiceCallProxy(object):
    # pylint: disable=R0903
    """
    Service call proxy
    """

    def __init__(self, uid, name, topic_prefix, publish_method):
        """
        Sets up the call proxy

        :param uid: End point UID
        :param name: End point name
        :param topic_prefix: Prefix for MQTT topics
        :param publish_method: Method to call to publish the request message
        """
        self.__uid = uid
        self.__name = name
        self.__topic_prefix = topic_prefix
        self.__publish = publish_method

    def __getattr__(self, name):
        """
        Prefixes the requested attribute name by the endpoint name
        """
        # Make a proxy for this call
        # This is an ugly trick to handle multi-threaded calls, as the
        # underlying proxy re-uses the same connection when possible: sometimes
        # it means sending a request before retrieving a result
        return _MqttCallableProxy(
            self.__uid,
            self.__topic_prefix,
            "{0}.{1}".format(self.__name, name),
            self.__publish,
        )


@ComponentFactory(pelix.remote.FACTORY_TRANSPORT_MQTTRPC_IMPORTER)
@Provides(pelix.remote.SERVICE_IMPORT_ENDPOINT_LISTENER)
@Property(
    "_kinds",
    pelix.remote.PROP_REMOTE_CONFIGS_SUPPORTED,
    (MQTTRPC_CONFIGURATION,),
)
@Property("_host", "mqtt.host", "localhost")
@Property("_port", "mqtt.port", 1883)
class MqttRpcServiceImporter(commons.AbstractRpcServiceImporter):
    """
    MQTT-RPC Remote Services importer
    """

    def __init__(self):
        """
        Sets up the exporter
        """
        # Call parent
        super(MqttRpcServiceImporter, self).__init__()

        # Component properties
        self._kinds = None

        # MQTT server
        self._host = None
        self._port = None

        # MQTT client
        self.__mqtt = None

        # Proxies waiting for an answer (correlation ID -> _MqttCallableProxy)
        self.__waiting = {}

        # Endpoints in use: Endpoint UID -> _MqttCallableProxy
        self.__waiting_endpoints = {}

    def make_service_proxy(self, endpoint):
        """
        Creates the proxy for the given ImportEndpoint

        :param endpoint: An ImportEndpoint bean
        :return: A service proxy
        """
        # Get the request topic
        topic_prefix = endpoint.properties.get(PROP_MQTT_TOPIC)
        if not topic_prefix:
            # No topic information
            _logger.warning("No MQTT topic given: %s", endpoint)
            return None

        return _ServiceCallProxy(
            endpoint.uid, endpoint.name, topic_prefix, self.__send_request
        )

    def clear_service_proxy(self, endpoint):
        """
        Destroys the proxy made for the given ImportEndpoint

        :param endpoint: An ImportEndpoint bean
        """
        try:
            # Free waiting calls
            for proxy in self.__waiting_endpoints.pop(endpoint.uid):
                proxy.handle_result(None, "Endpoint removed")

        except KeyError:
            # Unused endpoint
            return

    def __send_request(
        self, endpoint_uid, proxy, topic_prefix, request_parameters
    ):
        """
        Sends a request to the given topic

        :param endpoint_uid: Endpoint UID
        :param proxy: Callable proxy to notify on response
        :param topic_prefix: Prefix of MQTT topics to use
        :param request_parameters: Call parameters
        """
        # Prepare the correlation ID
        correlation_id = str(uuid.uuid4())

        try:
            # Prepare the data
            request = json.dumps(
                {
                    KEY_DATA: request_parameters,
                    KEY_CORRELATION_ID: correlation_id,
                    KEY_SENDER: self._framework_uid,
                }
            )

        except ValueError as ex:
            raise RemoteServiceError(
                "Cannot convert request to JSON: {0}".format(ex)
            )

        # Keep the callable in the waiting list
        self.__waiting[correlation_id] = proxy
        self.__waiting_endpoints.setdefault(endpoint_uid, set()).add(proxy)

        # Subscribe to the reply
        self.__mqtt.subscribe(make_topic(topic_prefix, TOPIC_RESPONSE))

        # Send the request
        self.__mqtt.publish(
            make_topic(topic_prefix, TOPIC_REQUEST), request, qos=2
        )

    def __on_message(self, client, msg):
        # pylint: disable=W0613
        """
        An MQTT reply has been received
        """
        try:
            # Parse data
            data = json.loads(to_str(msg.payload))

            # Check if we are the sender
            try:
                if data[KEY_SENDER] == self._framework_uid:
                    # We published this message
                    return
            except KeyError:
                # Not sent by us
                pass

            # Extract the correlation ID
            correlation_id = data[KEY_CORRELATION_ID]
        except ValueError as ex:
            # Unreadable reply
            _logger.error("Error reading MQTT-RPC reply: %s", ex)
            return
        except KeyError as ex:
            # No correlation ID
            _logger.error("Incomplete MQTT-RPC reply: missing %s", ex)
            return

        try:
            # Extract result
            result = data[KEY_DATA]
            error = data[KEY_ERROR]
        except KeyError as ex:
            # Incomplete result
            result = None
            error = "Missing MQTT-RPC reply field: {0}".format(ex)

        try:
            # Find the matching proxy
            proxy = self.__waiting.pop(correlation_id)
        except KeyError:
            # No a correlation ID we know
            pass
        else:
            # Notify the proxy
            proxy.handle_result(result, error)

    @Validate
    def validate(self, context):
        """
        Component validated
        """
        # Call the parent
        super(MqttRpcServiceImporter, self).validate(context)

        # Create the MQTT client
        self.__mqtt = pelix.misc.mqtt_client.MqttClient()

        # Customize callbacks
        self.__mqtt.on_message = self.__on_message

        # Prepare the connection
        self.__mqtt.connect(self._host, self._port)

    @Invalidate
    def invalidate(self, context):
        """
        Component invalidated
        """
        # Disconnect from the server (this stops the loop)
        self.__mqtt.disconnect()

        # Unlock proxies
        for proxy in self.__waiting.values():
            proxy.handle_result(None, "MQTT-RPC Importer stopped")

        # Clean up the storage
        self.__waiting.clear()

        # Call the parent
        super(MqttRpcServiceImporter, self).invalidate(context)

        # Clean up members
        self.__mqtt = None



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix remote services: Zeroconf (mDNS) discovery and event notification

This module depends on the zeroconf package

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import json
import logging
import socket

# Zeroconf
import zeroconf

# iPOPO decorators
from pelix.ipopo.decorators import (
    ComponentFactory,
    Requires,
    Provides,
    Invalidate,
    Validate,
    Property,
)
import pelix.constants

# Remote services
import pelix.remote
import pelix.remote.beans as beans
from pelix.utilities import is_bytes, is_string, to_str

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


@ComponentFactory(pelix.remote.FACTORY_DISCOVERY_ZEROCONF)
@Provides(pelix.remote.SERVICE_EXPORT_ENDPOINT_LISTENER)
@Property("_rs_type", pelix.remote.PROP_ZEROCONF_TYPE, "_pelix-rs._tcp.local.")
@Property("_ttl", "zeroconf.ttl", 60)
@Requires("_access", pelix.remote.SERVICE_DISPATCHER_SERVLET)
@Requires("_registry", pelix.remote.SERVICE_REGISTRY)
class ZeroconfDiscovery(object):
    """
    Remote services discovery and notification using the module zeroconf
    """

    # Service type for the Pelix dispatcher servlet
    DNS_DISPATCHER_TYPE = "_rs-dispatcher._tcp.local."

    def __init__(self):
        """
        Sets up the component
        """
        # Imported endpoints registry
        self._registry = None

        # Dispatcher access
        self._access = None

        # Remote Service type
        self._rs_type = None

        # Zeroconf TTL
        self._ttl = 60

        # Framework UID
        self._fw_uid = None

        # Address of this framework
        self._address = None

        # Zeroconf
        self._zeroconf = None  # type: zeroconf.Zeroconf
        self._browsers = []

        # Endpoint UID -> ServiceInfo
        self._export_infos = {}

        # mDNS name -> Endpoint UID
        self._imported_endpoints = {}

    @Invalidate
    def invalidate(self, _):
        """
        Component invalidated
        """
        # Stop listeners
        for browser in self._browsers:
            browser.cancel()

        # Close Zeroconf
        self._zeroconf.unregister_all_services()
        self._zeroconf.close()

        # Clean up
        self._export_infos.clear()
        self._zeroconf = None
        self._fw_uid = None

        _logger.debug("Zeroconf discovery invalidated")

    @Validate
    def validate(self, context):
        """
        Component validated
        """
        # Get the framework UID
        self._fw_uid = context.get_property(pelix.constants.FRAMEWORK_UID)

        # Get the host address
        self._address = socket.inet_aton(
            socket.gethostbyname(socket.gethostname())
        )

        # Prepare Zeroconf
        self._zeroconf = zeroconf.Zeroconf()

        # Register the dispatcher servlet as a service
        self.__register_servlet()

        # Listen to our types
        self._browsers.append(
            zeroconf.ServiceBrowser(
                self._zeroconf, ZeroconfDiscovery.DNS_DISPATCHER_TYPE, self
            )
        )
        self._browsers.append(
            zeroconf.ServiceBrowser(self._zeroconf, self._rs_type, self)
        )

        _logger.debug("Zeroconf discovery validated")

    @staticmethod
    def _serialize_properties(props):
        """
        Converts properties values into strings
        """
        new_props = {}

        for key, value in props.items():
            if is_string(value):
                new_props[key] = value
            else:
                try:
                    new_props[key] = json.dumps(value)
                except ValueError:
                    new_props[key] = "pelix-type:{0}:{1}".format(
                        type(value).__name__, repr(value)
                    )

        # FIXME: to simplify the usage with ECF, send single strings instead of
        # arrays
        for key in (
            pelix.constants.OBJECTCLASS,
            pelix.remote.PROP_IMPORTED_CONFIGS,
        ):
            try:
                new_props[key] = props[key][0]
            except KeyError:
                pass

        return new_props

    @staticmethod
    def _deserialize_properties(props):
        """
        Converts properties values into their type
        """
        new_props = {}
        for key, value in props.items():
            key = to_str(key)

            if is_bytes(value):
                # Convert value to string if necessary
                value = to_str(value)

            try:
                try:
                    new_props[key] = json.loads(value)
                except (TypeError, ValueError):
                    if is_string(value) and value.startswith("pelix-type:"):
                        # Pseudo-serialized
                        value_type, value = value.split(":", 3)[2:]
                        if "." in value_type and value_type not in value:
                            # Not a builtin type...
                            _logger.warning(
                                "Won't work: %s (%s)", value, value_type
                            )

                        new_props[key] = eval(value)
                    else:
                        # String
                        new_props[key] = value
            except Exception as ex:
                _logger.error("Can't deserialize %s: %s", value, ex)

        return new_props

    def __register_servlet(self):
        """
        Registers the Pelix Remote Services dispatcher servlet as a service via
        mDNS
        """
        # Get the dispatcher servlet access
        access = self._access.get_access()

        # Convert properties to be stored as strings
        properties = {
            "pelix.version": pelix.__version__,
            pelix.remote.PROP_ENDPOINT_FRAMEWORK_UUID: self._fw_uid,
            "pelix.access.port": access[0],
            "pelix.access.path": access[1],
        }
        properties = self._serialize_properties(properties)

        # Prepare the service type
        svc_name = "{0}.{1}".format(
            self._fw_uid, ZeroconfDiscovery.DNS_DISPATCHER_TYPE
        )

        # Prepare the mDNS entry
        info = zeroconf.ServiceInfo(
            ZeroconfDiscovery.DNS_DISPATCHER_TYPE,  # Type
            svc_name,  # Name
            self._address,  # Access address
            access[0],  # Access port
            properties=properties,
        )

        # Register the service
        self._zeroconf.register_service(info, self._ttl)

    def endpoints_added(self, endpoints):
        """
        Multiple endpoints have been added

        :param endpoints: A list of ExportEndpoint beans
        """
        # Get the dispatcher servlet port
        access_port = self._access.get_access()[0]

        # Handle each one separately
        for endpoint in endpoints:
            self._endpoint_added(endpoint, access_port)

    def _endpoint_added(self, exp_endpoint, access_port):
        """
        A new service is exported

        :param exp_endpoint: An ExportEndpoint bean
        :param access_port: The dispatcher access port
        """
        # Convert the export endpoint into an EndpointDescription bean
        endpoint = beans.EndpointDescription.from_export(exp_endpoint)

        # Get its properties
        properties = endpoint.get_properties()

        # Convert properties to be stored as strings
        properties = self._serialize_properties(properties)

        # Prepare the service name
        svc_name = "{0}.{1}".format(
            endpoint.get_id().replace("-", ""), self._rs_type
        )

        # Prepare the mDNS entry
        info = zeroconf.ServiceInfo(
            self._rs_type,  # Type
            svc_name,  # Name
            self._address,  # Access address
            access_port,  # Access port
            properties=properties,
        )

        self._export_infos[exp_endpoint.uid] = info

        # Register the service
        self._zeroconf.register_service(info, self._ttl)

    @staticmethod
    def endpoint_updated(endpoint, old_properties):
        # pylint: disable=W0613
        """
        An end point is updated

        :param endpoint: The updated endpoint
        :param old_properties: Previous properties of the endpoint
        """
        # Not available...
        # TODO: register a temporary service while the update is performed ?
        return

    def endpoint_removed(self, endpoint):
        """
        An end point is removed

        :param endpoint: Endpoint being removed
        """
        try:
            # Get the associated service info
            info = self._export_infos.pop(endpoint.uid)
        except KeyError:
            # Unknown service
            _logger.debug("Unknown removed endpoint: %s", endpoint)
        else:
            # Unregister the service
            self._zeroconf.unregister_service(info)

    def _get_service_info(self, svc_type, name, max_retries=10):
        # type: (str, str, int) -> zeroconf.ServiceInfo
        """
        Tries to get information about the given mDNS service

        :param svc_type: Service type
        :param name: Service name
        :param max_retries: Number of retries before timeout
        :return: A ServiceInfo bean
        """
        info = None
        retries = 0
        while (
            self._zeroconf is not None
            and info is None
            and retries < max_retries
        ):
            # Try to get information about the service...
            info = self._zeroconf.get_service_info(svc_type, name)
            retries += 1

        return info

    def add_service(self, zeroconf_, svc_type, name):
        """
        Called by Zeroconf when a record is updated

        :param zeroconf_: The Zeroconf instance than notifies of the
                          modification
        :param svc_type: Service type
        :param name: Service name
        """
        # Get information about the service
        info = self._get_service_info(svc_type, name)
        if info is None:
            _logger.warning(
                "Timeout reading service information: %s - %s", svc_type, name
            )
            return

        # Read properties
        properties = self._deserialize_properties(info.properties)

        try:
            sender_uid = properties[pelix.remote.PROP_ENDPOINT_FRAMEWORK_UUID]
            if sender_uid == self._fw_uid:
                # We sent this message
                return
        except KeyError:
            # Not a Pelix message
            _logger.warning("Not a Pelix record: %s", properties)
            return

        if svc_type == ZeroconfDiscovery.DNS_DISPATCHER_TYPE:
            # Dispatcher servlet found, get source info
            address = to_str(socket.inet_ntoa(info.address))
            port = info.port
            self._access.send_discovered(
                address, port, properties["pelix.access.path"]
            )
        elif svc_type == self._rs_type:
            # Remote service
            # Get the first available configuration
            configuration = properties[pelix.remote.PROP_IMPORTED_CONFIGS]
            if not is_string(configuration):
                configuration = configuration[0]

            # Ensure we have a list of specifications
            specs = properties[pelix.constants.OBJECTCLASS]
            if is_string(specs):
                specs = [specs]

            try:
                # Make an import bean
                endpoint = beans.ImportEndpoint(
                    properties[pelix.remote.PROP_ENDPOINT_ID],
                    properties[pelix.remote.PROP_ENDPOINT_FRAMEWORK_UUID],
                    [configuration],
                    None,
                    specs,
                    properties,
                )
            except KeyError as ex:
                # Log a warning on incomplete endpoints
                _logger.warning(
                    "Incomplete endpoint description, missing %s: %s",
                    ex,
                    properties,
                )
                return
            else:
                # Register the endpoint
                if self._registry.add(endpoint):
                    # Associate the mDNS name to the endpoint on success
                    self._imported_endpoints[name] = endpoint.uid

    def remove_service(self, zeroconf_, svc_type, name):
        """
        Called by Zeroconf when a record is removed

        :param zeroconf_: The Zeroconf instance than notifies of the
                          modification
        :param svc_type: Service type
        :param name: Service name
        """
        if svc_type == self._rs_type:
            # Get information about the service
            try:
                # Get the stored endpoint UID
                uid = self._imported_endpoints.pop(name)
            except KeyError:
                # Unknown service
                return
            else:
                # Remove it
                self._registry.remove(uid)
        elif svc_type == ZeroconfDiscovery.DNS_DISPATCHER_TYPE:
            # A dispatcher servlet is gone
            fw_uid = name.split(".", 1)[0]
            if fw_uid == self._fw_uid:
                # Local message: ignore
                return

            # Remote framework is lost
            self._registry.lost_framework(fw_uid)



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix remote services: Imported end points registry

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging
import threading

# Remote Services constants
import pelix.constants
import pelix.remote
import pelix.remote.beans as beans

# iPOPO decorators
from pelix.ipopo.decorators import (
    ComponentFactory,
    Requires,
    Provides,
    Instantiate,
    Invalidate,
    Validate,
    BindField,
)

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


@ComponentFactory("pelix-remote-imports-registry-factory")
@Provides(pelix.remote.SERVICE_REGISTRY)
@Requires(
    "_listeners",
    pelix.remote.SERVICE_IMPORT_ENDPOINT_LISTENER,
    aggregate=True,
    optional=True,
)
@Instantiate("pelix-remote-imports-registry")
class ImportsRegistry(object):
    """
    Registry of discovered end points. End points are identified by their UID
    """

    def __init__(self):
        # Listeners (injected)
        self._listeners = []

        # Framework UID
        self._fw_uid = None

        # Framework UID -> [ImportEndpoint]
        self._frameworks = {}

        # End point UID -> ImportEndpoint
        self._registry = {}

        # Lock
        self.__lock = threading.Lock()

    @BindField("_listeners", if_valid=True)
    def _bind_listener(self, field, listener, svc_ref):
        # pylint: disable=W0613
        """
        New listener bound
        """
        with self.__lock:
            # Late listener
            for endpoint in self._registry.values():
                try:
                    listener.endpoint_added(endpoint)
                except Exception as ex:
                    _logger.exception("Error calling listener: %s", ex)

    def add(self, endpoint):
        """
        Registers an end point and notifies listeners. Does nothing if the
        endpoint UID was already known.

        :param endpoint: An :class:`~pelix.remote.beans.ImportEndpoint` object
        :return: True if the end point has been added
        """
        with self.__lock:
            # Check framework UID (avoid to import our own services)
            if endpoint.framework == self._fw_uid:
                return False

            # Check if the end point already exists
            if endpoint.uid in self._registry:
                # Already known end point: do nothing
                _logger.debug("Already known endpoint")
                return False

            # Store the end point
            self._registry[endpoint.uid] = endpoint
            if endpoint.framework:
                self._frameworks.setdefault(endpoint.framework, []).append(
                    endpoint
                )

        # Notify listeners (out of lock)
        if self._listeners:
            for listener in self._listeners[:]:
                try:
                    listener.endpoint_added(endpoint)
                except Exception as ex:
                    _logger.exception("Error calling listener: %s", ex)
        return True

    def update(self, uid, new_properties):
        """
        Updates an end point and notifies listeners

        :param uid: The UID of the end point
        :param new_properties: The new properties of the end point
        :return: True if the endpoint is known, else False
        """
        try:
            with self.__lock:
                # Update the stored end point
                stored_endpoint = self._registry[uid]

                # Replace the stored properties
                old_properties = stored_endpoint.properties.copy()
                stored_endpoint.properties = new_properties
        except KeyError:
            # Unknown end point: ignore it
            return False
        else:
            # Notify listeners
            if self._listeners:
                for listener in self._listeners[:]:
                    try:
                        listener.endpoint_updated(
                            stored_endpoint, old_properties
                        )
                    except Exception as ex:
                        _logger.exception("Error calling listener: %s", ex)
            return True

    def contains(self, endpoint):
        """
        Checks if an endpoint is in the registry

        :param endpoint: An endpoint UID or an
                         :class:`~pelix.remote.beans.ImportEndpoint` object
        :return: True if the endpoint is known, else False
        """
        if isinstance(endpoint, beans.ImportEndpoint):
            return endpoint.uid in self._registry

        return endpoint in self._registry

    # Support for the 'in' keyword
    __contains__ = contains

    def remove(self, uid):
        """
        Unregisters an end point and notifies listeners

        :param uid: The UID of the end point to unregister
        :return: True if the endpoint was known
        """
        # Remove the end point from the individual storage
        try:
            endpoint = self._registry.pop(uid)
        except KeyError:
            # Unknown end point
            _logger.debug("Unknown end point UID: %s", uid)
            return False

        # Remove it from its framework storage, if any
        try:
            framework_endpoints = self._frameworks[endpoint.framework]
            if endpoint in framework_endpoints:
                framework_endpoints.remove(endpoint)
                if not framework_endpoints:
                    # Remove framework entry if there is no more endpoint
                    # from it
                    del self._frameworks[endpoint.framework]
        except (KeyError, ValueError):
            # Ignore the absence of reference in the framework storage
            pass

        # Notify listeners
        if self._listeners:
            for listener in self._listeners[:]:
                try:
                    listener.endpoint_removed(endpoint)
                except Exception as ex:
                    _logger.exception("Error calling listener: %s", ex)

        return True

    def lost_framework(self, uid):
        """
        Unregisters all the end points associated to the given framework UID

        :param uid: The UID of a framework
        """
        # Get the end points of this framework
        endpoints = self._frameworks.pop(uid, [])
        for endpoint in endpoints:
            with self.__lock:
                # Remove endpoint from registry
                try:
                    del self._registry[endpoint.uid]
                except KeyError:
                    # The endpoint may have been removed by a listener
                    pass

            # Notify listeners
            if self._listeners:
                for listener in self._listeners[:]:
                    try:
                        listener.endpoint_removed(endpoint)
                    except Exception as ex:
                        _logger.exception("Error calling listener: %s", ex)

    @Validate
    def _validate(self, context):
        """
        Component validated
        """
        # Get the framework UID
        self._fw_uid = context.get_property(pelix.constants.FRAMEWORK_UID)

    @Invalidate
    def _invalidate(self, _):
        """
        Component invalidated: clean up storage
        """
        # Clean up
        self._fw_uid = None
        self._frameworks.clear()
        self._registry.clear()



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
iPOPO installation script

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import os

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


def read(fname):
    """
    Utility method to read the content of a whole file
    """
    with open(os.path.join(os.path.dirname(__file__), fname)) as fd:
        return fd.read()


setup(
    name="iPOPO",
    version=__version__,
    license="Apache License 2.0",
    license_file="LICENSE",
    description="A service-oriented component model framework",
    long_description=read("README.rst"),
    author="Thomas Calmant",
    author_email="thomas.calmant@gmail.com",
    url="https://github.com/tcalmant/ipopo/",
    packages=[
        "pelix",
        "pelix.http",
        "pelix.internals",
        "pelix.ipopo",
        "pelix.ipopo.handlers",
        "pelix.misc",
        "pelix.remote",
        "pelix.remote.discovery",
        "pelix.remote.transport",
        "pelix.rsa",
        "pelix.rsa.providers",
        "pelix.rsa.providers.discovery",
        "pelix.rsa.providers.distribution",
        "pelix.rsa.topologymanagers",
        "pelix.services",
        "pelix.shell",
        "pelix.shell.completion",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
    install_requires=["jsonrpclib-pelix>=0.3.1"],
    extras_require={
        "Redis": ["redis>=2.10"],
        "ZooKeeper": ["kazoo>=2.4"],
        "MQTT": ["paho-mqtt>=1.3"],
        "XMPP": ["sleekxmpp>=1.3.1"],
        "zeroconf": ["zeroconf==0.19"],
        "RSA": ["python-etcd>=0.4.5", "osgiservicebridge>=1.5.1"],
    },
)



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix remote services: MQTT discovery provider

A discovery packet contains JSON representation of an ImportEndpoint bean.
This module depends on the paho-mqtt package (ex-mosquitto), provided by the
Eclipse Foundation: see http://www.eclipse.org/paho

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging

# MQTT client
import pelix.misc.mqtt_client

# iPOPO decorators
from pelix.ipopo.decorators import (
    ComponentFactory,
    Requires,
    Provides,
    Validate,
    Property,
    Invalidate,
)

# Pelix & Remote services
from pelix.remote.edef_io import EDEFWriter, EDEFReader
import pelix.constants as constants
import pelix.remote
import pelix.remote.beans as beans

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

_logger = logging.getLogger(__name__)

EVENT_ADD = "add"
EVENT_UPDATE = "update"
EVENT_REMOVE = "remove"
EVENT_LOST = "lost"
EVENT_DISCOVER = "discover"

ENDPOINT_EVENTS = (EVENT_ADD, EVENT_UPDATE, EVENT_REMOVE)

# ------------------------------------------------------------------------------


@ComponentFactory(pelix.remote.FACTORY_DISCOVERY_MQTT)
@Provides(pelix.remote.SERVICE_EXPORT_ENDPOINT_LISTENER, "_controller")
@Requires("_dispatcher", pelix.remote.SERVICE_DISPATCHER)
@Requires("_registry", pelix.remote.SERVICE_REGISTRY)
@Property("_host", "mqtt.host", "localhost")
@Property("_port", "mqtt.port", 1883)
@Property("_prefix", "topic.prefix", "pelix/{appid}/remote-services")
@Property("_appid", "application.id", None)
class MqttDiscovery(object):
    """
    Remote Service discovery provider based on MQTT
    """

    def __init__(self):
        """
        Sets up members
        """
        # Imports registry
        self._registry = None

        # Exports registry
        self._dispatcher = None

        # Service controller
        self._controller = False

        # Framework UID
        self._framework_uid = None

        # MQTT server properties
        self._host = "localhost"
        self._port = 1883

        # MQTT topic Properties
        self._prefix = ""
        self._appid = None

        # MQTT client
        self.__mqtt = None

        # Real prefix
        self._real_prefix = ""

    @Validate
    def _validate(self, context):
        """
        Component validated
        """
        # Format the topic prefix
        self._real_prefix = self._prefix.format(appid=self._appid or "")

        # Avoid double slashes
        self._real_prefix = self._real_prefix.replace("//", "/")

        # Get the framework UID
        self._framework_uid = context.get_property(constants.FRAMEWORK_UID)

        # Create the MQTT client
        self.__mqtt = pelix.misc.mqtt_client.MqttClient()

        # Customize callbacks
        self.__mqtt.on_connect = self.__on_connect
        self.__mqtt.on_disconnect = self.__on_disconnect
        self.__mqtt.on_message = self.__on_message

        # Prepare the will packet
        self.__mqtt.set_will(
            self._make_topic(EVENT_LOST), self._framework_uid, qos=2
        )

        # Prepare the connection
        self.__mqtt.connect(self._host, self._port)

    @Invalidate
    def _invalidate(self, _):
        """
        Component invalidated
        """
        # Send the "lost" message
        mid = self.__send_message(EVENT_LOST, self._framework_uid, True)
        self.__mqtt.wait_publication(mid, 10)

        # Disconnect from the server (this stops the loop)
        self.__mqtt.disconnect()

        # Clean up
        self._framework_uid = None
        self.__mqtt = None

    def _make_topic(self, event):
        """
        Prepares a MQTT topic name for the given event

        :param event: An event type (add, update, remove)
        :return: A MQTT topic
        """
        return "{0}/{1}".format(self._real_prefix, event)

    def __on_connect(self, client, result_code):
        """
        Client connected to the server
        """
        if not result_code:
            # Connection is OK, subscribe to the topic
            client.subscribe(self._make_topic("#"))

            # Provide the service
            self._controller = True

            # Send a discovery packet
            self.__send_message(EVENT_DISCOVER, self._framework_uid)

    def __on_disconnect(self, client, result_code):
        # pylint: disable=W0613
        """
        Client has been disconnected from the server
        """
        # Disconnected: stop providing the service
        self._controller = False

    def __on_message(self, client, msg):
        # pylint: disable=W0613
        """
        A message has been received from a server

        :param client: Client that received the message
        :param msg: A MQTTMessage bean
        """
        # Get the topic
        topic = msg.topic

        # Extract the event
        event = topic.rsplit("/", 1)[1]

        try:
            if event in ENDPOINT_EVENTS:
                # Parse the endpoints (from EDEF XML to ImportEndpoint)
                endpoints_descr = EDEFReader().parse(msg.payload)
                endpoints = [
                    endpoint.to_import() for endpoint in endpoints_descr
                ]

                if (
                    not endpoints
                    or endpoints[0].framework == self._framework_uid
                ):
                    # No enpoints to read or Loopback message
                    return

                # Give the list of endpoints to the handler
                parameter = endpoints
            else:
                # Give the payload as is to other event handlers
                parameter = msg.payload

            try:
                getattr(self, "_handle_{0}".format(event))(parameter)
            except AttributeError:
                _logger.error("Unhandled MQTT event: %s", event)

        except Exception as ex:
            _logger.exception(
                "Error handling an MQTT message '%s': %s", topic, ex
            )

    def __send_message(self, event, payload, wait=False):
        """
        Sends a message through the MQTT connection

        :param event: Remote service event name
        :param payload: Message content
        :return: The local message ID
        """
        # Publish the MQTT message (QoS 2 - Exactly Once)
        return self.__mqtt.publish(
            self._make_topic(event), payload, qos=2, wait=wait
        )

    def _handle_add(self, endpoints):
        """
        A set of endpoints have been registered

        :param endpoints: Parsed ImportEndpoint beans
        """
        # Notify the import registry
        for endpoint in endpoints:
            self._registry.add(endpoint)

    def _handle_update(self, endpoints):
        """
        A set of endpoints have been updated

        :param endpoints: Parsed ImportEndpoint beans
        """
        # Notify the import registry
        for endpoint in endpoints:
            self._registry.update(endpoint.uid, endpoint.properties)

    def _handle_remove(self, endpoints):
        """
        A set of endpoints has been removed

        :param endpoints: Parsed ImportEndpoint beans
        """
        # Notify the import registry
        for endpoint in endpoints:
            self._registry.remove(endpoint.uid)

    def _handle_discover(self, payload):
        """
        A framework wants to discover all services

        :param payload: The UID of the sender
        """
        if payload == self._framework_uid:
            # We are the sender, ignore this message
            return

        # Get the list of our exported services
        endpoints = self._dispatcher.get_endpoints()
        if not endpoints:
            # Nothing to say
            return

        # Convert the beans to XML (EDEF format)
        xml_string = EDEFWriter().to_string(
            beans.EndpointDescription.from_export(endpoint)
            for endpoint in endpoints
        )

        # Send the message
        self.__send_message(EVENT_ADD, xml_string)

    def _handle_lost(self, payload):
        """
        A framework has been lost

        :param payload: The UID of the lost framework
        """
        self._registry.lost_framework(payload)

    def endpoints_added(self, endpoints):
        """
        Multiple endpoints have been added

        :param endpoints: A list of ExportEndpoint beans
        """
        # Convert the beans to XML (EDEF format)
        xml_string = EDEFWriter().to_string(
            beans.EndpointDescription.from_export(endpoint)
            for endpoint in endpoints
        )

        # Send the message
        self.__send_message(EVENT_ADD, xml_string)

    def endpoint_updated(self, endpoint, old_properties):
        # pylint: disable=W0613
        """
        An end point is updated

        :param endpoint: The updated endpoint
        :param old_properties: Previous properties of the endpoint
        """
        # Convert the endpoint into an EndpointDescription bean
        endpoint_desc = beans.EndpointDescription.from_export(endpoint)

        # Convert the bean to XML (EDEF format)
        xml_string = EDEFWriter().to_string([endpoint_desc])

        # Send the message
        self.__send_message(EVENT_UPDATE, xml_string)

    def endpoint_removed(self, endpoint):
        """
        An end point is removed

        :param endpoint: Endpoint being removed
        """
        # Convert the endpoint into an EndpointDescription bean
        endpoint_desc = beans.EndpointDescription.from_export(endpoint)

        # Convert the bean to XML (EDEF format)
        xml_string = EDEFWriter().to_string([endpoint_desc])

        # Send the message
        self.__send_message(EVENT_REMOVE, xml_string)



#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""

EndpointDescription class API

:author: Scott Lewis
:copyright: Copyright 2020, Scott Lewis
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Scott Lewis

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import logging

# Typing
try:
    # pylint: disable=W0611
    from typing import Dict, Any, Optional, List, Tuple, Iterable
    from pelix.framework import ServiceReference
except ImportError:
    pass

from pelix.constants import SERVICE_ID, FRAMEWORK_UID, OBJECTCLASS
from pelix.ldapfilter import get_ldap_filter

from pelix.rsa import (
    set_prop_if_null,
    get_prop_value,
    get_matching_interfaces,
    ENDPOINT_SERVICE_ID,
    SERVICE_IMPORTED,
    ENDPOINT_FRAMEWORK_UUID,
    ENDPOINT_ID,
    ECF_ENDPOINT_ID,
    ECF_ENDPOINT_TIMESTAMP,
    ECF_ENDPOINT_CONNECTTARGET_ID,
    ECF_ENDPOINT_IDFILTER_IDS,
    ECF_RSVC_ID,
    ECF_ENDPOINT_CONTAINERID_NAMESPACE,
    ECF_ENDPOINT_REMOTESERVICE_FILTER,
    ECF_SERVICE_EXPORTED_ASYNC_INTERFACES,
    ECF_SERVICE_EXPORTED_ASYNC_NOPROXY,
    ECF_ASYNC_INTERFACE_SUFFIX,
    ECF_SERVICE_ASYNC_RSPROXY_CLASS_,
    ENDPOINT_PACKAGE_VERSION_,
    REMOTE_INTENTS_SUPPORTED,
    SERVICE_IMPORTED_CONFIGS,
    REMOTE_CONFIGS_SUPPORTED,
    SERVICE_INTENTS,
    is_reserved_property,
    merge_dicts,
    get_string_plus_property,
    rsid_to_string,
    OSGI_BASIC_TIMEOUT_INTENT,
    get_string_plus_property_value,
)

# ------------------------------------------------------------------------------
# Module version

__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# Standard logging
_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


def encode_list(key, list_):
    # type: (str, Iterable) -> Dict[str, str]
    """
    Converts a list into a space-separated string and puts it in a dictionary

    :param key: Dictionary key to store the list
    :param list_: A list of objects
    :return: A dictionary key->string or an empty dictionary
    """
    if not list_:
        return {}
    return {key: " ".join(str(i) for i in list_)}


def package_name(package):
    # type: (str) -> str
    """
    Returns the package name of the given module name
    """
    if not package:
        return ""

    lastdot = package.rfind(".")
    if lastdot == -1:
        return package

    return package[:lastdot]


def encode_osgi_props(ed):
    # type: (EndpointDescription) -> Dict[str, str]
    """
    Prepares a dictionary of OSGi properties for the given EndpointDescription
    """
    result_props = {}
    intfs = ed.get_interfaces()
    result_props[OBJECTCLASS] = " ".join(intfs)
    for intf in intfs:
        pkg_name = package_name(intf)
        ver = ed.get_package_version(pkg_name)
        if ver and not ver == (0, 0, 0):
            result_props[ENDPOINT_PACKAGE_VERSION_] = ".".join(
                str(v) for v in ver
            )

    result_props[ENDPOINT_ID] = ed.get_id()
    result_props[ENDPOINT_SERVICE_ID] = "{0}".format(ed.get_service_id())
    result_props[ENDPOINT_FRAMEWORK_UUID] = ed.get_framework_uuid()
    imp_configs = ed.get_imported_configs()
    if imp_configs:
        result_props[SERVICE_IMPORTED_CONFIGS] = " ".join(
            ed.get_imported_configs()
        )
    intents = ed.get_intents()
    if intents:
        result_props[SERVICE_INTENTS] = " ".join(intents)
    remote_configs = ed.get_remote_configs_supported()
    if remote_configs:
        result_props[REMOTE_CONFIGS_SUPPORTED] = " ".join(remote_configs)
    remote_intents = ed.get_remote_intents_supported()
    if remote_intents:
        result_props[REMOTE_INTENTS_SUPPORTED] = " ".join(remote_intents)
    return result_props


def decode_list(input_props, name):
    # type: (Dict[str, str], str) -> List[str]
    """
    Decodes a space-separated list
    """
    val_str = input_props.get(name, None)
    if val_str:
        return val_str.split(" ")
    return []


def decode_osgi_props(input_props):
    # type: (Dict[str, Any]) -> Dict[str, Any]
    """
    Decodes the OSGi properties of the given endpoint properties
    """
    result_props = {}
    intfs = decode_list(input_props, OBJECTCLASS)
    result_props[OBJECTCLASS] = intfs
    for intf in intfs:
        package_key = ENDPOINT_PACKAGE_VERSION_ + package_name(intf)
        intfversionstr = input_props.get(package_key, None)
        if intfversionstr:
            result_props[package_key] = intfversionstr
    result_props[ENDPOINT_ID] = input_props[ENDPOINT_ID]
    result_props[ENDPOINT_SERVICE_ID] = input_props[ENDPOINT_SERVICE_ID]
    result_props[ENDPOINT_FRAMEWORK_UUID] = input_props[ENDPOINT_FRAMEWORK_UUID]
    imp_configs = decode_list(input_props, SERVICE_IMPORTED_CONFIGS)
    if imp_configs:
        result_props[SERVICE_IMPORTED_CONFIGS] = imp_configs
    intents = decode_list(input_props, SERVICE_INTENTS)
    if intents:
        result_props[SERVICE_INTENTS] = intents
    remote_configs = decode_list(input_props, REMOTE_CONFIGS_SUPPORTED)
    if remote_configs:
        result_props[REMOTE_CONFIGS_SUPPORTED] = remote_configs
    remote_intents = decode_list(input_props, REMOTE_INTENTS_SUPPORTED)
    if remote_intents:
        result_props[REMOTE_INTENTS_SUPPORTED] = remote_intents
    return result_props


def decode_endpoint_props(input_props):
    # type: (Dict) -> Dict[str, Any]
    """
    Decodes the endpoint properties from the given dictionary
    """
    ed_props = decode_osgi_props(input_props)
    ed_props[ECF_ENDPOINT_CONTAINERID_NAMESPACE] = input_props[
        ECF_ENDPOINT_CONTAINERID_NAMESPACE
    ]
    ed_props[ECF_RSVC_ID] = int(input_props[ECF_RSVC_ID])
    ed_props[ECF_ENDPOINT_ID] = input_props[ECF_ENDPOINT_ID]
    ed_props[ECF_ENDPOINT_TIMESTAMP] = int(input_props[ECF_ENDPOINT_TIMESTAMP])
    target_id = input_props.get(ECF_ENDPOINT_CONNECTTARGET_ID, None)
    if target_id:
        ed_props[ECF_ENDPOINT_CONNECTTARGET_ID] = target_id
    id_filters = decode_list(input_props, ECF_ENDPOINT_IDFILTER_IDS)
    if id_filters:
        ed_props[ECF_ENDPOINT_IDFILTER_IDS] = id_filters
    rs_filter = input_props.get(ECF_ENDPOINT_REMOTESERVICE_FILTER, None)
    if rs_filter:
        ed_props[ECF_ENDPOINT_REMOTESERVICE_FILTER] = rs_filter
    async_intfs = input_props.get(ECF_SERVICE_EXPORTED_ASYNC_INTERFACES, None)
    if async_intfs:
        if async_intfs == "*":
            ed_props[ECF_SERVICE_EXPORTED_ASYNC_INTERFACES] = async_intfs
        else:
            async_intfs = decode_list(
                input_props, ECF_SERVICE_EXPORTED_ASYNC_INTERFACES
            )
            if async_intfs:
                ed_props[ECF_SERVICE_EXPORTED_ASYNC_INTERFACES] = async_intfs

    for key in input_props.keys():
        if not is_reserved_property(key):
            val = input_props.get(key, None)
            if val:
                ed_props[key] = val
    return ed_props


def encode_endpoint_props(ed):
    """
    Encodes the properties of the given EndpointDescription
    """
    props = encode_osgi_props(ed)
    props[ECF_RSVC_ID] = "{0}".format(ed.get_remoteservice_id()[1])
    props[ECF_ENDPOINT_ID] = "{0}".format(ed.get_container_id()[1])
    props[ECF_ENDPOINT_CONTAINERID_NAMESPACE] = "{0}".format(
        ed.get_container_id()[0]
    )
    props[ECF_ENDPOINT_TIMESTAMP] = "{0}".format(ed.get_timestamp())
    ctid = ed.get_connect_target_id()
    if ctid:
        props[ECF_ENDPOINT_CONNECTTARGET_ID] = "{0}".format(ctid)
    id_filters = ed.get_id_filters()
    if id_filters:
        props[ECF_ENDPOINT_IDFILTER_IDS] = " ".join([x[1] for x in id_filters])
    rs_filter = ed.get_remoteservice_filter()
    if rs_filter:
        props[ECF_ENDPOINT_REMOTESERVICE_FILTER] = ed.get_remoteservice_filter()
    async_intfs = ed.get_async_interfaces()
    if async_intfs:
        props[ECF_SERVICE_EXPORTED_ASYNC_INTERFACES] = " ".join(async_intfs)

    all_props = ed.get_properties()
    other_props = {
        key: all_props[key]
        for key in all_props.keys()
        if not is_reserved_property(key)
    }
    return merge_dicts(props, other_props)


# ------------------------------------------------------------------------------
# EndpointDescription class
# ------------------------------------------------------------------------------


class EndpointDescription(object):
    """
    Description of an RSA Endpoint
    """

    @classmethod
    def fromsvcref(cls, svc_ref):
        # type: (ServiceReference) -> EndpointDescription
        return cls(svc_ref, None)

    @classmethod
    def fromprops(cls, props):
        # type: (Dict[str, Any]) -> EndpointDescription
        return cls(None, props)

    @classmethod
    def fromsvcrefprops(cls, svc_ref, props):
        # type: (ServiceReference, Dict[str, Any]) -> EndpointDescription
        return cls(svc_ref, props)

    @classmethod
    def _condition_props(cls, properties):
        # type: (Dict[str, Any]) -> Dict[str, Any]
        set_prop_if_null(SERVICE_IMPORTED, properties, True)
        for key in properties.keys():
            if key.startswith("services.exported."):
                del properties[key]
        return properties

    @classmethod
    def _verify_export_props(cls, svc_ref, all_properties):
        # type: (ServiceReference, Dict[str, Any]) -> Dict[str, Any]
        props = all_properties.copy()
        set_prop_if_null(
            ENDPOINT_SERVICE_ID, props, svc_ref.get_property(SERVICE_ID)
        )
        set_prop_if_null(
            ENDPOINT_FRAMEWORK_UUID, props, svc_ref.get_property(FRAMEWORK_UID)
        )
        return props

    def __init__(self, svc_ref=None, properties=None):
        # type: (Optional[ServiceReference], Optional[Dict[str, Any]]) -> None
        if svc_ref is None and properties is None:
            raise ValueError(
                "Either service reference or properties argument must be "
                "non-null"
            )

        all_properties = {}  # type: Dict[str, Any]

        if svc_ref is not None:
            all_properties.update(svc_ref.get_properties())

        if properties is not None:
            all_properties.update(properties)

        if svc_ref is not None:
            self._properties = EndpointDescription._verify_export_props(
                svc_ref, all_properties
            )
        else:
            self._properties = all_properties

        self._interfaces = self._properties.get(OBJECTCLASS)[:]
        self._service_id = self._verify_long_prop(ENDPOINT_SERVICE_ID)
        self._framework_uuid = self._verify_str_prop(ENDPOINT_FRAMEWORK_UUID)
        endpoint_id = self._verify_str_prop(ENDPOINT_ID)
        if endpoint_id is None:
            raise ValueError("endpoint.id property must be set")
        self._id = endpoint_id.strip()

        if not self.get_configuration_types():
            raise ValueError(
                "service.imported.configs property must be set and non-empty"
            )

        self._ecfid = self._verify_str_prop(ECF_ENDPOINT_ID)
        if self._ecfid is None:
            raise ValueError("ecf.endpoint.id must not be null")
        self._timestamp = self._verify_long_prop(ECF_ENDPOINT_TIMESTAMP)
        if self._timestamp == 0:
            self._timestamp = self.get_service_id()
        self._id_namespace = self._verify_str_prop(
            ECF_ENDPOINT_CONTAINERID_NAMESPACE
        )
        self._container_id = (self._id_namespace, self._ecfid)
        self._rs_id = self._verify_long_prop(ECF_RSVC_ID)
        if self._rs_id is None:
            self._rs_id = self.get_service_id()

        connect_target_name = self._get_prop(ECF_ENDPOINT_CONNECTTARGET_ID)
        if connect_target_name is not None:
            self._connect_target_id = (self._id_namespace, connect_target_name)
        else:
            self._connect_target_id = None

        id_filter_names = self._get_string_plus_property(
            ECF_ENDPOINT_IDFILTER_IDS
        )
        if id_filter_names:
            self._id_filters = [
                (self._id_namespace, x) for x in id_filter_names
            ]
        else:
            self._id_filters = None

        self._rs_filter = self._get_prop(ECF_ENDPOINT_REMOTESERVICE_FILTER)
        self._async_intfs = self._verify_async_intfs()

    def __hash__(self):
        return hash(self._id)

    def __eq__(self, other):
        # pylint: disable=W0212
        return self._id == other._id

    def __ne__(self, other):
        # pylint: disable=W0212
        return self._id != other._id

    def __str__(self):
        get_remoteservice_id = self.get_remoteservice_id()
        return (
            "EndpointDescription(id={0}; endpoint.service.id={1}; "
            "framework.uuid={2}; ecf.endpoint.id={3}:{4})".format(
                self.get_id(),
                self.get_service_id(),
                self.get_framework_uuid(),
                get_remoteservice_id[0],
                get_remoteservice_id[1],
            )
        )

    def _get_prop(self, key, default=None):
        return get_prop_value(key, self._properties, default)

    def _get_string_plus_property(self, key):
        return get_string_plus_property(key, self._properties, [])

    def _verify_long_prop(self, prop):
        value = self._get_prop(prop)
        return int(value) if value else int(0)

    def _verify_str_prop(self, prop):
        # type: (str) -> str
        value = self._get_prop(prop)
        if value is None:
            raise ValueError(
                "prop name=" + prop + " must be present in properties"
            )
        return str(value)

    def _convert_intf_to_async(self, intf):
        # type: (str) -> str
        async_proxy_intf = self._get_prop(ECF_SERVICE_ASYNC_RSPROXY_CLASS_)
        if async_proxy_intf is not None:
            return async_proxy_intf

        if intf.endswith(ECF_ASYNC_INTERFACE_SUFFIX):
            return intf

        return intf + ECF_ASYNC_INTERFACE_SUFFIX

    def _verify_async_intfs(self):
        # type: () -> List[str]
        matching = []  # type: List[str]
        no_async_prop = self._get_prop(ECF_SERVICE_EXPORTED_ASYNC_NOPROXY)
        if no_async_prop is None:
            async_inf_val = self._get_prop(
                ECF_SERVICE_EXPORTED_ASYNC_INTERFACES
            )
            if async_inf_val is not None:
                matching = get_matching_interfaces(
                    self.get_interfaces(), async_inf_val
                )
        return [self._convert_intf_to_async(x) for x in matching]

    def get_container_id(self):
        # type: () -> Tuple[str, str]
        return self._container_id

    def get_connect_target_id(self):
        # type: () -> Tuple[str, str]
        return self._connect_target_id

    def get_timestamp(self):
        # type: () -> int
        return self._timestamp

    def get_remoteservice_id(self):
        # type: () -> Tuple[Tuple[str, str], int]
        return (self.get_container_id(), self._rs_id)

    def get_remoteservice_idstr(self):
        # type: () -> str
        return rsid_to_string(self.get_remoteservice_id())

    def get_id_filters(self):
        # type: () -> Optional[List[Tuple[str, str]]]
        return self._id_filters

    def get_remoteservice_filter(self):
        # type: () -> Optional[str]
        return self._rs_filter

    def get_async_interfaces(self):
        # type: () -> List[str]
        return self._async_intfs

    def get_framework_uuid(self):
        # type: () -> str
        return self._framework_uuid

    def get_osgi_basic_timeout(self):
        # type: () -> Optional[int]
        timeout = self.get_properties().get(OSGI_BASIC_TIMEOUT_INTENT, None)
        if isinstance(timeout, str):
            timeout = int(timeout)
        return int(timeout / 1000) if timeout else None

    def get_id(self):
        # type: () -> str
        """
        Returns the endpoint's id.

        :return: str
        """
        return self._id

    def get_remote_intents_supported(self):
        # type: () -> List[str]
        return self._get_string_plus_property(REMOTE_INTENTS_SUPPORTED)

    def get_intents(self):
        # type: () -> List[str]
        """
        Returns the list of intents required by this endpoint.

        The intents are based on the service.intents on an imported service,
        except for any intents that are additionally provided by the importing
        distribution provider.
        All qualified intents must have been expanded.
        This value of the intents is stored in the
        rsa.SERVICE_INTENTS service property.

        :return: A list of intents (list of str)
        """
        # Return a copy of the list
        return self._get_string_plus_property(SERVICE_INTENTS)

    def get_interfaces(self):
        # type: () -> List[str]
        """
        Provides the list of interfaces implemented by the exported service.

        :return: A list of specifications (list of str)
        """
        return self._interfaces

    def get_imported_configs(self):
        # type: () -> List[str]
        return self.get_configuration_types()

    def update_imported_configs(self, imported_configs):
        self._properties[
            SERVICE_IMPORTED_CONFIGS
        ] = get_string_plus_property_value(imported_configs)

    def get_configuration_types(self):
        # type: () -> List[str]
        return self._get_string_plus_property(SERVICE_IMPORTED_CONFIGS)

    def get_remote_configs_supported(self):
        # type: () -> List[str]
        return self._get_string_plus_property(REMOTE_CONFIGS_SUPPORTED)

    def get_service_id(self):
        # type: () -> int
        return self._service_id

    def get_package_version(self, package):
        # type: (str) -> Tuple[int, int, int]
        """
        Provides the version of the given package name.

        :param package: The name of the package
        :return: The version of the specified package as a tuple or (0,0,0)
        """
        name = "{0}{1}".format(ENDPOINT_PACKAGE_VERSION_, package)
        try:
            # Get the version string
            version = self._properties[name]
            # Split dots ('.')
            return tuple(version.split("."))
        except KeyError:
            # No version
            return 0, 0, 0

    def get_properties(self):
        # type: () -> Dict[str, Any]
        """
        Returns all endpoint properties.

        :return: A copy of the endpoint properties
        """
        return self._properties.copy()

    def is_same_service(self, endpoint):
        # type: (EndpointDescription) -> bool
        """
        Tests if this endpoint and the given one have the same framework UUID
        and service ID

        :param endpoint: Another endpoint
        :return: True if both endpoints represent the same remote service
        """
        return (
            self.get_framework_uuid() == endpoint.get_framework_uuid()
            and self.get_service_id() == endpoint.get_service_id()
        )

    def matches(self, ldap_filter):
        # type: (str) -> bool
        """
        Tests the properties of this EndpointDescription against the given
        filter

        :param ldap_filter: A filter
        :return: True if properties matches the filter
        """
        return get_ldap_filter(ldap_filter).matches(self._properties)



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
RequiresMap handler implementation

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import copy
import logging
import threading

# Pelix beans
from pelix.constants import BundleActivator, BundleException
from pelix.internals.events import ServiceEvent

# iPOPO constants
import pelix.ipopo.constants as ipopo_constants
import pelix.ipopo.handlers.constants as constants

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


class _HandlerFactory(constants.HandlerFactory):
    # pylint: disable=R0903
    """
    Factory service for service registration handlers
    """

    @staticmethod
    def _prepare_requirements(configs, requires_filters):
        """
        Overrides the filters specified in the decorator with the given ones

        :param configs: Field → (Requirement, key, allow_none) dictionary
        :param requires_filters: Content of the 'requires.filter' component
                                 property (field → string)
        :return: The new configuration dictionary
        """
        if not requires_filters or not isinstance(requires_filters, dict):
            # No explicit filter configured
            return configs

        # We need to change a part of the requirements
        new_requirements = {}
        for field, config in configs.items():
            # Extract values from tuple
            requirement, key, allow_none = config

            try:
                explicit_filter = requires_filters[field]

                # Store an updated copy of the requirement
                requirement_copy = requirement.copy()
                requirement_copy.set_filter(explicit_filter)
                new_requirements[field] = (requirement_copy, key, allow_none)

            except (KeyError, TypeError, ValueError):
                # No information for this one, or invalid filter:
                # keep the factory requirement
                new_requirements[field] = config

        return new_requirements

    def get_handlers(self, component_context, instance):
        """
        Sets up service providers for the given component

        :param component_context: The ComponentContext bean
        :param instance: The component instance
        :return: The list of handlers associated to the given component
        """
        # Extract information from the context
        configs = component_context.get_handler(
            ipopo_constants.HANDLER_REQUIRES_MAP
        )
        requires_filters = component_context.properties.get(
            ipopo_constants.IPOPO_REQUIRES_FILTERS, None
        )

        # Prepare requirements
        configs = self._prepare_requirements(configs, requires_filters)

        # Set up the runtime dependency handlers
        handlers = []
        for field, config in configs.items():
            # Extract values from tuple
            requirement, key, allow_none = config

            # Construct the handler
            if requirement.aggregate:
                handlers.append(
                    AggregateDependency(field, requirement, key, allow_none)
                )
            else:
                handlers.append(
                    SimpleDependency(field, requirement, key, allow_none)
                )

        return handlers


@BundleActivator
class _Activator(object):
    """
    The bundle activator
    """

    def __init__(self):
        """
        Sets up members
        """
        self._registration = None

    def start(self, context):
        """
        Bundle started
        """
        # Set up properties
        properties = {
            constants.PROP_HANDLER_ID: ipopo_constants.HANDLER_REQUIRES_MAP
        }

        # Register the handler factory service
        self._registration = context.register_service(
            constants.SERVICE_IPOPO_HANDLER_FACTORY,
            _HandlerFactory(),
            properties,
        )

    def stop(self, _):
        """
        Bundle stopped
        """
        # Unregister the service
        self._registration.unregister()
        self._registration = None


# ------------------------------------------------------------------------------


class _RuntimeDependency(constants.DependencyHandler):
    """
    Manages a required dependency field when a component is running
    """

    def __init__(self, field, requirement, key, allow_none):
        """
        Sets up the dependency

        :param field: The injected field name
        :param requirement: The Requirement describing this dependency
        :param key: The property used as key in the dictionary
        :param allow_none: Allow None property as key
        """
        # The internal state lock
        self._lock = threading.RLock()

        # The iPOPO StoredInstance object (given during manipulation)
        self._ipopo_instance = None

        # The bundle context
        self._context = None

        # The associated field
        self._field = field

        # The underlying requirement
        self.requirement = requirement

        # The property name
        self._key = key

        # Accept None values
        self._allow_none = allow_none

        # Reference -> Service
        self.services = {}

        # Future injected dictionary
        self._future_value = {}

    def manipulate(self, stored_instance, component_instance):
        """
        Stores the given StoredInstance bean.

        :param stored_instance: The iPOPO component StoredInstance
        :param component_instance: The component instance
        """
        # Store the stored instance...
        self._ipopo_instance = stored_instance

        # ... and the bundle context
        self._context = stored_instance.bundle_context

        # Set the default value for the field: an empty dictionary
        setattr(component_instance, self._field, {})

    def clear(self):
        """
        Cleans up the manager. The manager can't be used after this method has
        been called
        """
        self.services.clear()
        self._future_value.clear()

        self.services = None
        self._lock = None
        self._ipopo_instance = None
        self._context = None
        self.requirement = None
        self._key = None
        self._allow_none = None
        self._future_value = None
        self._field = None

    def get_bindings(self):
        """
        Retrieves the list of the references to the bound services

        :return: A list of ServiceReferences objects
        """
        with self._lock:
            return list(self.services.keys())

    def get_field(self):
        """
        Returns the name of the field handled by this handler
        """
        return self._field

    def get_kinds(self):
        """
        Retrieves the kinds of this handler: 'dependency'

        :return: the kinds of this handler
        """
        return (constants.KIND_DEPENDENCY,)

    def get_value(self):
        """
        Retrieves the value to inject in the component

        :return: The value to inject
        """
        # Return a copy of the future value
        with self._lock:
            # IronPython can't copy dictionary with a None key
            return copy.copy(self._future_value)

    def is_valid(self):
        """
        Tests if the dependency is in a valid state
        """
        return (
            self.requirement is not None and self.requirement.optional
        ) or bool(self._future_value)

    def on_service_arrival(self, svc_ref):
        """
        Called when a service has been registered in the framework

        :param svc_ref: A service reference
        """
        raise NotImplementedError

    def on_service_departure(self, svc_ref):
        """
        Called when a service has been registered in the framework

        :param svc_ref: A service reference
        """
        raise NotImplementedError

    def on_service_modify(self, svc_ref, old_properties):
        """
        Called when a service has been registered in the framework

        :param svc_ref: A service reference
        :param old_properties: Previous properties values
        """
        raise NotImplementedError

    def service_changed(self, event):
        """
        Called by the framework when a service event occurs
        """
        if (
            self._ipopo_instance is None
            or not self._ipopo_instance.check_event(event)
        ):
            # stop() and clean() may have been called after we have been put
            # inside a listener list copy...
            # or we've been told to ignore this event
            return

        # Call sub-methods
        kind = event.get_kind()
        svc_ref = event.get_service_reference()

        if kind == ServiceEvent.REGISTERED:
            # Service coming
            self.on_service_arrival(svc_ref)

        elif kind in (
            ServiceEvent.UNREGISTERING,
            ServiceEvent.MODIFIED_ENDMATCH,
        ):
            # Service gone or not matching anymore
            self.on_service_departure(svc_ref)

        elif kind == ServiceEvent.MODIFIED:
            # Modified properties (can be a new injection)
            self.on_service_modify(svc_ref, event.get_previous_properties())

    def start(self):
        """
        Starts the dependency manager
        """
        self._context.add_service_listener(
            self, self.requirement.filter, self.requirement.specification
        )

    def stop(self):
        """
        Stops the dependency manager (must be called before clear())

        :return: The removed bindings (list) or None
        """
        self._context.remove_service_listener(self)
        if self.services:
            return [
                (service, reference)
                for reference, service in self.services.items()
            ]

        return None

    def try_binding(self):
        """
        Searches for the required service if needed

        :raise BundleException: Invalid ServiceReference found
        """
        with self._lock:
            if self.services:
                # We already are alive (not our first call)
                # => we are updated through service events
                return

            # Get all matching services
            refs = self._context.get_all_service_references(
                self.requirement.specification, self.requirement.filter
            )
            if not refs:
                # No match found
                return

            results = []
            try:
                # Bind all new reference
                for reference in refs:
                    added = self.on_service_arrival(reference)
                    if added:
                        results.append(reference)
            except BundleException as ex:
                # Get the logger for this instance
                logger = logging.getLogger(
                    "-".join((self._ipopo_instance.name, "RequiresMap-Runtime"))
                )
                logger.debug("Error binding multiple references: %s", ex)

                # Undo what has just been done, ignoring errors
                for reference in results:
                    try:
                        self.on_service_departure(reference)

                    except BundleException as ex2:
                        logger.debug("Error cleaning up: %s", ex2)

                del results[:]
                raise


class SimpleDependency(_RuntimeDependency):
    """
    Manages a simple dependency field: one service per dictionary key
    """

    def on_service_arrival(self, svc_ref):
        """
        Called when a service has been registered in the framework

        :param svc_ref: A service reference
        """
        with self._lock:
            if svc_ref not in self.services:
                # Get the key property
                prop_value = svc_ref.get_property(self._key)
                if (
                    prop_value not in self._future_value
                    and prop_value is not None
                    or self._allow_none
                ):
                    # Matching new property value
                    service = self._context.get_service(svc_ref)

                    # Store the information
                    self._future_value[prop_value] = service
                    self.services[svc_ref] = service

                    # Call back iPOPO
                    self._ipopo_instance.bind(self, service, svc_ref)
                    return True

            return None

    def on_service_departure(self, svc_ref):
        """
        Called when a service has been unregistered from the framework

        :param svc_ref: A service reference
        """
        with self._lock:
            if svc_ref in self.services:
                # Get the service instance
                service = self.services.pop(svc_ref)

                # Get the key property
                prop_value = svc_ref.get_property(self._key)

                # Remove the injected service
                del self._future_value[prop_value]

                self._ipopo_instance.unbind(self, service, svc_ref)
                return True

            return None

    def on_service_modify(self, svc_ref, old_properties):
        """
        Called when a service has been modified in the framework

        :param svc_ref: A service reference
        :param old_properties: Previous properties values
        """
        with self._lock:
            if svc_ref not in self.services:
                # A previously registered service now matches our filter
                return self.on_service_arrival(svc_ref)
            else:
                # Get the property values
                old_value = old_properties.get(self._key)
                prop_value = svc_ref.get_property(self._key)
                service = self.services[svc_ref]

                if old_value != prop_value:
                    if (
                        prop_value is not None
                        or self._allow_none
                        and prop_value not in self._future_value
                    ):
                        # New property accepted and not yet in use
                        del self._future_value[old_value]
                        self._future_value[prop_value] = service

                        # Notify the property modification, with a value change
                        self._ipopo_instance.update(
                            self, service, svc_ref, old_properties, True
                        )
                    else:
                        # Consider the service as gone
                        del self._future_value[old_value]
                        del self.services[svc_ref]
                        self._ipopo_instance.unbind(self, service, svc_ref)
                else:
                    # Notify the property modification
                    self._ipopo_instance.update(
                        self, service, svc_ref, old_properties, False
                    )

            return None


class AggregateDependency(_RuntimeDependency):
    """
    Manages an aggregated dependency field: multiple services per dictionary
    key
    """

    def __store_service(self, key, service):
        """
        Stores the given service in the dictionary

        :param key: Dictionary key
        :param service: Service to add to the dictionary
        """
        self._future_value.setdefault(key, []).append(service)

    def __remove_service(self, key, service):
        """
        Removes the given service from the future dictionary

        :param key: Dictionary key
        :param service: Service to remove from the dictionary
        """
        try:
            # Remove the injected service
            prop_services = self._future_value[key]
            prop_services.remove(service)

            # Clean up
            if not prop_services:
                del self._future_value[key]

        except KeyError:
            # Ignore: can occur when removing a service with a None property,
            # if allow_none is False
            pass

    def get_value(self):
        """
        Retrieves the value to inject in the component

        :return: The value to inject
        """
        with self._lock:
            # The value field must be a deep copy of our dictionary
            if self._future_value is not None:
                return {
                    key: value[:] for key, value in self._future_value.items()
                }

            return None

    def on_service_arrival(self, svc_ref):
        """
        Called when a service has been registered in the framework

        :param svc_ref: A service reference
        :return: True if the service is consumed
        """
        with self._lock:
            if svc_ref not in self.services:
                # Get the key property
                prop_value = svc_ref.get_property(self._key)
                if prop_value is not None or self._allow_none:
                    # Get the new service
                    service = self._context.get_service(svc_ref)

                    # Store the information
                    self.__store_service(prop_value, service)
                    self.services[svc_ref] = service

                    self._ipopo_instance.bind(self, service, svc_ref)
                    return True

            return None

    def on_service_departure(self, svc_ref):
        """
        Called when a service has been unregistered from the framework

        :param svc_ref: A service reference
        :return: A tuple (service, reference) if the service has been lost,
                 else None
        """
        with self._lock:
            if svc_ref in self.services:
                # Get the service instance
                service = self.services.pop(svc_ref)

                # Get the key property
                prop_value = svc_ref.get_property(self._key)

                # Remove the injected service
                self.__remove_service(prop_value, service)

                self._ipopo_instance.unbind(self, service, svc_ref)
                return True

            return None

    def on_service_modify(self, svc_ref, old_properties):
        """
        Called when a service has been modified in the framework

        :param svc_ref: A service reference
        :param old_properties: Previous properties values
        :return: A tuple (added, (service, reference)) if the dependency has
                 been changed, else None
        """
        with self._lock:
            if svc_ref not in self.services:
                # A previously registered service now matches our filter
                return self.on_service_arrival(svc_ref)
            else:
                # Get the property values
                service = self.services[svc_ref]
                old_value = old_properties.get(self._key)
                prop_value = svc_ref.get_property(self._key)

                if old_value != prop_value:
                    # Key changed
                    if prop_value is not None or self._allow_none:
                        # New property accepted
                        if old_value is not None or self._allow_none:
                            self.__remove_service(old_value, service)

                        self.__store_service(prop_value, service)

                        # Notify the property modification, with a value change
                        self._ipopo_instance.update(
                            self, service, svc_ref, old_properties, True
                        )
                    else:
                        # Consider the service as gone
                        self.__remove_service(old_value, service)
                        del self.services[svc_ref]
                        self._ipopo_instance.unbind(self, service, svc_ref)
                else:
                    # Simple property update
                    self._ipopo_instance.update(
                        self, service, svc_ref, old_properties, False
                    )

            return None



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix HTTP package.

Defines the interfaces that must respect HTTP service implementations.

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard typing module should be optional
try:
    # pylint: disable=W0611
    from typing import Any, ByteString, Dict, Iterable, IO, Tuple
except ImportError:
    pass

# Pelix utility methods
from pelix.utilities import to_bytes

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

# HTTP service constants
HTTP_SERVICE = "pelix.http.service"
""" HTTP Service specification """

# ... binding address
HTTP_SERVICE_ADDRESS = "pelix.http.address"
""" HTTP Service binding address property (string) """

# ... binding port
HTTP_SERVICE_PORT = "pelix.http.port"
""" HTTP Service binding port property (int) """

# ... server uses SSL (read-only flag)
HTTP_USES_SSL = "pelix.https"
""" Read-only flag indicating if the server is using SSL (HTTPS) """

# ... the certificate file for HTTPS servers
HTTPS_CERT_FILE = "pelix.https.cert_file"
""" Path to the certificate file to configure a HTTPS server """

# ... the key file for HTTPS servers
HTTPS_KEY_FILE = "pelix.https.key_file"
""" Path to the certificate key file to configure a HTTPS server """

# ... the password of the key file for HTTPS servers
# (supported since Python 3.3)
HTTPS_KEY_PASSWORD = "pelix.https.key_password"

# HTTP servlet constants
HTTP_SERVLET = "pelix.http.servlet"
""" HTTP Servlet service specification """

# ... servlet path(s) (string or list of strings)
HTTP_SERVLET_PATH = "pelix.http.path"
""" HTTP Servlet path(s) (string or list or tuple of strings) """

# Service to provide custom 404 and 500 error pages
HTTP_ERROR_PAGES = "pelix.http.error.pages"

# ------------------------------------------------------------------------------

FACTORY_HTTP_BASIC = "pelix.http.service.basic.factory"
""" Name of the HTTP service component factory """

# ------------------------------------------------------------------------------

PARAM_NAME = "http.name"
"""
Entry in the parameters dictionary of ``bound_to`` and ``unbound_from``.
Contains the name of the server.
If the HTTP service is implemented with iPOPO, it might be the instance name.
"""

PARAM_EXTRA = "http.extra"
"""
Entry in the parameters dictionary of ``bound_to`` and ``unbound_from``.
Contains a copy of extra properties of the HTTP service implementation.
Its content is implementation dependent.
"""

PARAM_ADDRESS = "http.address"
"""
Entry in the parameters dictionary of ``bound_to`` and ``unbound_from``.
Contains the socket binding address of the HTTP server binding the servlet
"""

PARAM_PORT = "http.port"
"""
Entry in the parameters dictionary of ``bound_to`` and ``unbound_from``.
Contains the listening port of the HTTP server binding the servlet
"""

PARAM_HTTPS = "http.https"
"""
Entry in the parameters dictionary of ``bound_to`` and ``unbound_from``.
Contains a boolean: if True, the connection to the server is encrypted (HTTPS)
"""

# ------------------------------------------------------------------------------


def make_html_list(items, tag="ul"):
    # type: (Iterable[Any], str) -> str
    """
    Makes a HTML list from the given iterable

    :param items: The items to list
    :param tag: The tag to use (ul or ol)
    :return: The HTML list code
    """
    html_list = "\n".join(
        '<li><a href="{0}">{0}</a></li>'.format(item) for item in items
    )
    return "<{0}>\n{1}\n</{0}>".format(tag, html_list)


# ------------------------------------------------------------------------------


class AbstractHTTPServletRequest(object):
    """
    Abstract HTTP Servlet request helper
    """

    def get_command(self):
        # type: () -> str
        """
        Returns the HTTP verb (GET, POST, ...) used for the request
        """
        raise NotImplementedError(
            "This method must be implemented by a child class"
        )

    def get_client_address(self):
        # type: () -> Tuple[str, int]
        """
        Returns the address of the client

        :return: A (host, port) tuple
        """
        raise NotImplementedError(
            "This method must be implemented by a child class"
        )

    def get_header(self, name, default=None):
        # type: (str, Any) -> Any
        """
        Returns the value of a header

        :param name: Header name
        :param default: Default value if the header doesn't exist
        :return: The header value or the default one
        """
        raise NotImplementedError(
            "This method must be implemented by a child class"
        )

    def get_headers(self):
        # type: () -> Dict[str, Any]
        """
        Returns a copy all headers, with a dictionary interface

        :return: A dictionary-like object
        """
        raise NotImplementedError(
            "This method must be implemented by a child class"
        )

    def get_path(self):
        # type: () -> str
        """
        Returns the request full path

        :return: A request full path (string)
        """
        raise NotImplementedError(
            "This method must be implemented by a child class"
        )

    def get_prefix_path(self):
        # type: () -> str
        """
        Returns the path to the servlet root

        :return: A request path (string)
        """
        raise NotImplementedError(
            "This method must be implemented by a child class"
        )

    def get_sub_path(self):
        # type: () -> str
        """
        Returns the servlet-relative path, i.e. after the prefix

        :return: A request path (string)
        """
        raise NotImplementedError(
            "This method must be implemented by a child class"
        )

    def get_rfile(self):
        # type: () -> IO
        """
        Returns the request input as a file stream

        :return: A file-like input stream
        """
        raise NotImplementedError(
            "This method must be implemented by a child class"
        )

    def read_data(self):
        # type: () -> ByteString
        """
        Reads all the data in the input stream

        :return: The read data
        """
        try:
            size = int(self.get_header("content-length"))
        except (ValueError, TypeError):
            size = -1

        return self.get_rfile().read(size)


class AbstractHTTPServletResponse(object):
    """
    HTTP Servlet response helper
    """

    def set_response(self, code, message=None):
        # type: (int, str) -> None
        """
        Sets the response line.
        This method should be the first called when sending an answer.

        :param code: HTTP result code
        :param message: Associated message
        """
        raise NotImplementedError(
            "This method must be implemented by a child class"
        )

    def set_header(self, name, value):
        # type: (str, Any) -> None
        """
        Sets the value of a header.
        This method should not be called after ``end_headers()``.

        :param name: Header name
        :param value: Header value
        """
        raise NotImplementedError(
            "This method must be implemented by a child class"
        )

    def is_header_set(self, name):
        # type: (str) -> bool
        """
        Checks if the given header has already been set

        :param name: Header name
        :return: True if it has already been set
        """
        raise NotImplementedError(
            "This method must be implemented by a child class"
        )

    def end_headers(self):
        """
        Ends the headers part
        """
        raise NotImplementedError(
            "This method must be implemented by a child class"
        )

    def get_wfile(self):
        # type: () -> IO
        """
        Retrieves the output as a file stream.
        ``end_headers()`` should have been called before, except if you want
        to write your own headers.

        :return: A file-like output stream
        """
        raise NotImplementedError(
            "This method must be implemented by a child class"
        )

    def write(self, data):
        # type: (ByteString) -> None
        """
        Writes the given data.
        ``end_headers()`` should have been called before, except if you want
        to write your own headers.

        :param data: Data to be written
        """
        raise NotImplementedError(
            "This method must be implemented by a child class"
        )

    def send_content(
        self,
        http_code,
        content,
        mime_type="text/html",
        http_message=None,
        content_length=-1,
    ):
        # type: (int, str, str, str, int) -> None
        """
        Utility method to send the given content as an answer.
        You can still use get_wfile or write afterwards, if you forced the
        content length.

        If content_length is negative (default), it will be computed as the
        length of the content;
        if it is positive, the given value will be used;
        if it is None, the content-length header won't be sent.

        :param http_code: HTTP result code
        :param content: Data to be sent (must be a string)
        :param mime_type: Content MIME type (content-type)
        :param http_message: HTTP code description
        :param content_length: Forced content length
        """
        self.set_response(http_code, http_message)
        if mime_type and not self.is_header_set("content-type"):
            self.set_header("content-type", mime_type)

        # Convert the content
        raw_content = to_bytes(content)

        if content_length is not None and not self.is_header_set(
            "content-length"
        ):
            if content_length < 0:
                # Compute the length
                content_length = len(raw_content)

            # Send the length
            self.set_header("content-length", content_length)

        self.end_headers()

        # Send the content
        self.write(raw_content)
